"""
A utility function to be called when optimizing observation configurations
"""
import logging
import os
from pprint import pformat

import optuna
import yaml
from optuna.integration.skopt import SkoptSampler
from optuna.pruners import SuccessiveHalvingPruner, MedianPruner
from optuna.samplers import RandomSampler, TPESampler
from stable_baselines.common.base_class import _UnvecWrapper
from stable_baselines.common.vec_env import VecEnv
from stable_baselines.her import HERGoalEnvWrapper

from commonroad_rl.utils_run.callbacks import ObservationConfigsTrialEvalCallback

LOGGER = logging.getLogger(__name__)


def optimize_observation_configs(
        algo,
        env_id,
        model_fn,
        env_fn,
        sampling_setting,
        n_timesteps=5000,
        eval_freq=1000,
        n_eval_episodes=5,
        n_trials=10,
        hyperparams=None,
        configs=None,
        n_jobs=1,
        sampler_method="random",
        pruner_method="halving",
        seed=13,
        verbose=1,
        log_path=None,
        best_model_save_path=None,
):
    """

    :param algo: (str)
    :param env: (str)
    :param model_fn: (func) function that is used to instantiate the model
    :param env_fn: (func) function that is used to instantiate the env
    :param n_trials: (int) maximum number of trials for finding the best hyperparams
    :param n_timesteps: (int) maximum number of timesteps per trial
    :param hyperparams: (dict) model hyperparameters
    :param configs: (dict) environment configurations to be optimized
    :param n_jobs: (int) number of parallel jobs
    :param sampler_method: (str)
    :param pruner_method: (str)
    :param seed: (int)
    :param sampling_setting: (dict) sampling intervals for correspinding items
    :param verbose: (int)
    :param log_path: (str) folder for saving evaluation results during optimization
    :param best_model_save_path: (str) folder for saving the best model
    :return: (dict) detailed result of the optimization
    """
    LOGGER.setLevel(verbose)

    if not len(LOGGER.handlers):
        formatter = logging.Formatter("[%(levelname)s] %(name)s - %(message)s")
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(verbose)
        stream_handler.setFormatter(formatter)
        LOGGER.addHandler(stream_handler)

        if log_path is not None:
            file_handler = logging.FileHandler(filename=os.path.join(log_path, "console_copy.txt"))
            file_handler.setLevel(verbose)
            file_handler.setFormatter(formatter)
            LOGGER.addHandler(file_handler)

    LOGGER.info(f"Optimizing observation configurations for {env_id}")

    # Enviornment configurations including observation space settings and reward weights
    if configs is None:
        configs = {}

    n_startup_trials = 10
    n_evaluations = n_timesteps // eval_freq

    LOGGER.info(f"Optimizing with {n_trials} trials using {n_jobs} parallel jobs, "
                f"each with {n_timesteps} maximal time steps and {n_evaluations} evaluations")

    # n_warmup_steps: Disable pruner until the trial reaches the given number of step.
    if sampler_method == "random":
        sampler = RandomSampler(seed=seed)
    elif sampler_method == "tpe":
        sampler = TPESampler(n_startup_trials=n_startup_trials, seed=seed)
    elif sampler_method == "skopt":
        # cf https://scikit-optimize.github.io/#skopt.Optimizer
        # GP: gaussian process
        # Gradient boosted regression: GBRT
        sampler = SkoptSampler(skopt_kwargs={"base_estimator": "GP", "acq_func": "gp_hedge"})
    else:
        raise ValueError(f"[observation_configs_opt.py] Unknown sampler: {sampler_method}")

    if pruner_method == "halving":
        pruner = SuccessiveHalvingPruner(min_resource=1, reduction_factor=4, min_early_stopping_rate=0)
    elif pruner_method == "median":
        pruner = MedianPruner(n_startup_trials=n_startup_trials, n_warmup_steps=n_evaluations // 3)
    elif pruner_method == "none":
        # Do not prune
        pruner = MedianPruner(n_startup_trials=n_trials, n_warmup_steps=n_evaluations)
    else:
        raise ValueError(f"[observation_configs_opt.py] Unknown pruner: {pruner_method}")

    LOGGER.debug(f"Sampler: {sampler_method} - Pruner: {pruner_method}")

    # Create study object on environment configurations from Optuna
    observation_configs_study = optuna.create_study(sampler=sampler, pruner=pruner)

    # Prepare the sampler
    observation_configs_sampler = OBSERVATION_CONFIGS_SAMPLER[env_id]

    def objective_observation_configs(trial):
        """
        Optimization objective for environment configurations
        """
        trial.model_class = None

        if algo == "her":
            trial.model_class = hyperparams["model_class"]

        # Hack to use DDPG/TD3 noise sampler
        if algo in ["ddpg", "td3"] or trial.model_class in ["ddpg", "td3"]:
            trial.n_actions = env_fn(n_envs=1).action_space.shape[0]

        # Get keyword arguments for model hyperparameters and environment configurations
        kwargs_hyperparams = hyperparams.copy()
        kwargs_configs = configs.copy()

        # Sample environment configurations and keep model hyperparameters
        sampled_observation_configs = observation_configs_sampler(
            trial, sampling_setting
        )
        kwargs_configs.update(sampled_observation_configs)

        # Save data for later inspection
        tmp_path = os.path.join(log_path, "trial_" + str(trial.number))
        os.makedirs(tmp_path, exist_ok=True)
        with open(
                os.path.join(tmp_path, "sampled_observation_configurations.yml"), "w"
        ) as f:
            yaml.dump(kwargs_configs, f)
            LOGGER.info("Saving sampled observation configurations into " + tmp_path)
        LOGGER.debug("Sampled observation configurations:")
        LOGGER.debug(pformat(kwargs_configs))

        # Create model and environments for optimization
        model = model_fn(kwargs_hyperparams, kwargs_configs)
        eval_env = env_fn(n_envs=1, eval_env=True, **kwargs_configs)

        # Account for parallel envs
        eval_freq_ = eval_freq
        if isinstance(model.get_env(), VecEnv):
            eval_freq_ = max(eval_freq // model.get_env().num_envs, 1)

        LOGGER.info(
            "Evaluating with {} episodes after every {} time steps".format(
                n_eval_episodes, eval_freq_
            )
        )

        observation_configs_eval_callback = ObservationConfigsTrialEvalCallback(
            eval_env,
            trial,
            n_eval_episodes=n_eval_episodes,
            eval_freq=eval_freq_,
            log_path=log_path,
            best_model_save_path=best_model_save_path,
            deterministic=True,
            verbose=verbose,
        )

        if algo == "her":
            # Wrap the env if need to flatten the dict obs
            if isinstance(eval_env, VecEnv):
                eval_env = _UnvecWrapper(eval_env)
            eval_env = HERGoalEnvWrapper(eval_env)

        try:
            model.learn(n_timesteps, callback=observation_configs_eval_callback)
            # Free memory
            model.env.close()
            eval_env.close()
        except AssertionError:
            # Sometimes, random hyperparams can generate NaN
            # Free memory
            model.env.close()
            eval_env.close()
            raise optuna.exceptions.TrialPruned()

        is_pruned = observation_configs_eval_callback.is_pruned
        cost = observation_configs_eval_callback.cost
        del model.env, eval_env
        del model
        if is_pruned:
            raise optuna.exceptions.TrialPruned()
        return cost

    try:
        LOGGER.info(f"Trying to optimize observation configurations with {n_trials} trials and {n_jobs} jobs")
        observation_configs_study.optimize(objective_observation_configs, n_trials=n_trials, n_jobs=n_jobs)
    except KeyboardInterrupt:
        pass

    LOGGER.info(f"Number of finished trials: {len(observation_configs_study.trials)}")
    LOGGER.info(f"Best value: {observation_configs_study.best_trial.value}")
    LOGGER.info("Best observation configurations: ")
    LOGGER.info(pformat(observation_configs_study.best_trial.params))

    return observation_configs_study.best_trial.params


def sample_commonroad_observation_configs(trial, sampling_setting):
    observation_configs = {}
    for key, values in sampling_setting.items():
        if len(values) == 1:
            method, interval = next(iter(values.items()))
            if method == "categorical":
                observation_configs[key] = trial.suggest_categorical(key, interval)
            elif method == "uniform":
                observation_configs[key] = trial.suggest_uniform(key, interval[0], interval[1])
            elif method == "loguniform":
                observation_configs[key] = trial.suggest_loguniform(key, interval[0], interval[1])
            else:
                print(f"[observation_configs_opt.py] Sampling method {method} not supported")
        elif len(values) > 1:
            observation_configs[key] = dict()
            for subkey, subvalue in values.items():
                method, interval = next(iter(subvalue.items()))
                if method == "categorical":
                    observation_configs[key][subkey] = trial.suggest_categorical(subkey, interval)
                elif method == "uniform":
                    observation_configs[key][subkey] = trial.suggest_uniform(subkey, interval[0], interval[1])
                elif method == "loguniform":
                    observation_configs[key][subkey] = trial.suggest_loguniform(subkey, interval[0], interval[1])
                else:
                    print(f"[observation_configs_opt.py] Sampling method {method} not supported")

    return observation_configs


OBSERVATION_CONFIGS_SAMPLER = {
    "commonroad-v0": sample_commonroad_observation_configs,
    "commonroad-v1": sample_commonroad_observation_configs
}



import os
import glob
import ntpath
import argparse

from shutil import copyfile


def get_args():
    parser = argparse.ArgumentParser(description="Divide files into subfolders for mpi",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--input_dir", "-i", type=str, default="/data/highD-dataset-v1.0/cr_scenarios")
    parser.add_argument("--output_dir", "-o", type=str, default=None)
    parser.add_argument("--file_extension", "-f", type=str, default="*.xml")
    parser.add_argument("--n_cpus", "-n", type=int, default=1)
    parser.add_argument("--duplicate", "-d", action="store_true",
                        help="Duplicate scenarios to ensure every CPU gets some work (use for less scenarios than CPUs)")

    return parser.parse_args()


def main():
    args = get_args()

    fns = sorted(glob.glob(os.path.join(args.input_dir, args.file_extension)))
    # duplicate fns to match number of CPUs (roughly)
    while args.duplicate and len(fns) < args.n_cpus:
        fns = ((args.n_cpus + len(fns) - 1) // len(fns)) * fns  # round up to ensure there is enough work
    n_files_per_cpu = len(fns) // args.n_cpus

    if args.output_dir is None:
        args.output_dir = os.path.abspath(args.input_dir) + "_" + str(args.n_cpus)

    for n in range(args.n_cpus):
        subdir = os.path.join(args.output_dir, str(n))
        os.makedirs(subdir, exist_ok=True)
        for fn in fns[n * n_files_per_cpu: (n + 1) * n_files_per_cpu]:
            copyfile(fn, os.path.join(subdir, ntpath.basename(fn)))

    # Distribute the files that can't be evenly distributed
    if len(fns) > args.n_cpus * n_files_per_cpu:
        cpu_index = 0
        for fn in fns[args.n_cpus * n_files_per_cpu:]:
            copyfile(fn, os.path.join(args.output_dir, str(cpu_index), ntpath.basename(fn)))
            cpu_index = (cpu_index + 1) % args.n_cpus


if __name__ == "__main__":
    main()



"""Class for sparse reward"""
import logging

from commonroad_rl.gym_commonroad.action import Action
from commonroad_rl.gym_commonroad.reward.reward import Reward

LOGGER = logging.getLogger(__name__)


class SparseReward(Reward):
    """Class for sparse reward"""

    def __init__(self, configs: dict):
        self.reward_configs = configs["reward_configs"]["sparse_reward"]

    def calc_reward(self, observation_dict: dict, ego_action: Action) -> float:
        """
        Calculate the reward according to the observations

        :param observation_dict: current observations
        :param ego_action: Current ego_action of the environment
        :return: Reward of this step
        """
        reward = 0.0

        # Reach goal
        if observation_dict["is_goal_reached"][0]:
            LOGGER.debug("GOAL REACHED!")
            reward += self.reward_configs["reward_goal_reached"]
        # Collision
        if observation_dict["is_collision"][0]:
            reward += self.reward_configs["reward_collision"]
        # Off-road
        if observation_dict["is_off_road"][0]:
            reward += self.reward_configs["reward_off_road"]
        # Friction violation
        if observation_dict["is_friction_violation"][0]:
            reward += self.reward_configs["reward_friction_violation"]
        # Exceed maximum episode length
        if observation_dict["is_time_out"][0]:
            reward += self.reward_configs["reward_time_out"]

        return reward



import argparse
import glob
import importlib
import os
import re
import logging
LOGGER = logging.getLogger(__name__)

import gym
import yaml
import numpy as np
try:
    import pybullet_envs
except ImportError:
    pybullet_envs = None
try:
    import mpi4py
except ImportError:
    mpi4py = None

from stable_baselines.deepq.policies import FeedForwardPolicy
from stable_baselines.common.policies import FeedForwardPolicy as BasePolicy
from stable_baselines.common.policies import register_policy
from stable_baselines.sac.policies import FeedForwardPolicy as SACPolicy
from stable_baselines.bench import Monitor
from stable_baselines import logger
from stable_baselines import PPO2, A2C, ACER, ACKTR, HER, SAC, TD3, GAIL

# DDPG and TRPO require MPI to be installed
if mpi4py is None:
    DDPG, TRPO = None, None
else:
    from stable_baselines import DDPG, TRPO

from stable_baselines.common.vec_env import (
    DummyVecEnv,
    VecNormalize as SBVecNormalize,
    VecFrameStack,
    SubprocVecEnv,
)
from stable_baselines.common.cmd_util import make_atari_env
from stable_baselines.common import set_global_seeds, BaseRLModel

from commonroad_rl.gym_commonroad.constants import PATH_PARAMS
ALGOS = {
    "a2c": A2C,
    "acer": ACER,
    "acktr": ACKTR,
    "ddpg": DDPG,
    "her": HER,
    "sac": SAC,
    "ppo2": PPO2,
    "trpo": TRPO,
    "td3": TD3,
    "gail": GAIL,
}
# import DQN
try:
    from stable_baselines import DQN
except:
    print("Can't import DQN, skipped!")
else:
    ALGOS["dqn"] = DQN


# ================== Custom Policies =================


class CustomDQNPolicy(FeedForwardPolicy):
    def __init__(self, *args, **kwargs):
        super(CustomDQNPolicy, self).__init__(*args, **kwargs, layers=[64], layer_norm=True, feature_extraction="mlp")


class CustomMlpPolicy(BasePolicy):
    def __init__(self, *args, **kwargs):
        super(CustomMlpPolicy, self).__init__(*args, **kwargs, layers=[16], feature_extraction="mlp")


class CustomSACPolicy(SACPolicy):
    def __init__(self, *args, **kwargs):
        super(CustomSACPolicy, self).__init__(*args, **kwargs, layers=[256, 256], feature_extraction="mlp")


register_policy("CustomSACPolicy", CustomSACPolicy)
register_policy("CustomDQNPolicy", CustomDQNPolicy)
register_policy("CustomMlpPolicy", CustomMlpPolicy)


class CRVecNormalize(SBVecNormalize):
    def __init__(self, vec_normalize: SBVecNormalize):
        self.vec_normalize = vec_normalize
        # SBVecNormalize.__init__(self, SBVecNormalize)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(
                "attempted to get missing private attribute '{}'".format(name)
            )
        return getattr(self.vec_normalize, name)

    def reset(self, **kwargs):
        """
        Reset all environments
        """
        obs = self.venv.reset(**kwargs)
        self.old_obs = obs
        self.ret = np.zeros(self.num_envs)
        if self.training:
            self._update_reward(self.ret)
        return self.normalize_obs(obs)


def load_model_and_vecnormalize(model_path: str, algo: str, normalize: bool, env: gym.Env, model_name="best_model") -> BaseRLModel:
    """
    Load trained model and corresponding vecnormalize.pkl
    :param model_path: Path to folder containing the trained model
    :param algo: The used RL algorithm
    :param normalize: If the env was normalized during training
    :param env: The gym.Env used during training
    :return: best_model.zip if exists else the last model and corresponding VecNormalize wrapped Env
    """
    # Load the trained agent
    files = os.listdir(model_path)
    model_name = f"{model_name}.zip"
    if model_name in files:
        model_path = os.path.join(model_path, model_name)
        if normalize:
            vec_normalize_path = model_path.replace(model_name, "vecnormalize.pkl")
    else:
        # No best_model.zip, find last model
        files = sorted(glob.glob(os.path.join(model_path, "rl_model*.zip")))

        def extract_number(f):
            s = re.findall("\d+", f)
            return int(s[-1]) if s else -1, f

        model_path = max(files, key=extract_number)
        vec_normalize_path = model_path.replace("rl_model", "vecnormalize").replace(".zip", ".pkl")

    if os.path.exists(vec_normalize_path):
        LOGGER.info(f"Loading saved running average from {vec_normalize_path}")
        vec_normalize_env = SBVecNormalize.load(vec_normalize_path, env)
        vec_normalize_env.training = False
        env = CRVecNormalize(vec_normalize_env)
    else:
        raise FileNotFoundError(f"vecnormalize.pkl not found in {vec_normalize_path}")
        
    # During testing the vecnormalize should not update the moving average
    env.training = False
    LOGGER.info(f"Loading model from {model_path}")
    model = ALGOS[algo].load(model_path)

    return model, env


def flatten_dict_observations(env):
    assert isinstance(env.observation_space, gym.spaces.Dict)
    keys = env.observation_space.spaces.keys()
    return gym.wrappers.FlattenDictWrapper(env, dict_keys=list(keys))


def get_wrapper_class(hyperparams):
    """
    Get one or more Gym environment wrapper class specified as a hyper parameter
    "env_wrapper".
    e.g.
    env_wrapper: gym_minigrid.wrappers.FlatObsWrapper

    for multiple, specify a list:

    env_wrapper:
        - utils_run.wrappers.DoneOnSuccessWrapper:
            reward_offset: 1.0
        - utils_run.wrappers.TimeFeatureWrapper
        - utils_run.wrappers.IncreaseTimeStepWrapper:
            num_steps: 10


    :param hyperparams: (dict)
    :return: a subclass of gym.Wrapper (class object) you can use to
             create another Gym env giving an original env.
    """

    def get_module_name(wrapper_name):
        return ".".join(wrapper_name.split(".")[:-1])

    def get_class_name(wrapper_name):
        return wrapper_name.split(".")[-1]

    if "env_wrapper" in hyperparams.keys():
        wrapper_name = hyperparams.get("env_wrapper")

        if wrapper_name is None:
            return None

        if not isinstance(wrapper_name, list):
            wrapper_names = [wrapper_name]
        else:
            wrapper_names = wrapper_name

        wrapper_classes = []
        wrapper_kwargs = []
        # Handle multiple wrappers
        for wrapper_name in wrapper_names:
            # Handle keyword arguments
            if isinstance(wrapper_name, dict):
                assert len(wrapper_name) == 1
                wrapper_dict = wrapper_name
                wrapper_name = list(wrapper_dict.keys())[0]
                kwargs = wrapper_dict[wrapper_name]
            else:
                kwargs = {}
            wrapper_module = importlib.import_module(get_module_name(wrapper_name))
            wrapper_class = getattr(wrapper_module, get_class_name(wrapper_name))
            wrapper_classes.append(wrapper_class)
            wrapper_kwargs.append(kwargs)

        def wrap_env(env):
            """
            :param env: (gym.Env)
            :return: (gym.Env)
            """
            for wrapper_class, kwargs in zip(wrapper_classes, wrapper_kwargs):
                env = wrapper_class(env, **kwargs)
            return env

        return wrap_env
    else:
        return None


def make_env(
        env_id,
        rank=0,
        seed=0,
        log_dir=None,
        logging_path=None,
        wrapper_class=None,
        env_kwargs=None,
        subproc=False,
        info_keywords=(),
):
    """
    Helper function to multiprocess training
    and log the progress.

    :param env_id: (str)
    :param rank: (int)
    :param seed: (int)
    :param log_dir: (str) path to monitor information
    :param logging_path: (str) path to console_copy.txt
    :param wrapper_class: (type) a subclass of gym.Wrapper to wrap the original env with
    :param env_kwargs: (Dict[str, Any]) Optional keyword argument to pass to the env constructor
    :param subproc: (bool) Whether env is used in a SubprocVecEnv, used to specify scenario path
    :param info_keywords: (tuple) extra information to log, from the information return of environment.step
    """
    if log_dir is not None:
        os.makedirs(log_dir, exist_ok=True)

    if env_kwargs is None:
        env_kwargs = {}

    def _init():
        if env_id == "cr-monitor-v0":
            pass
        if "commonroad" in env_id:
            pass

        set_global_seeds(seed + rank)
        if subproc and ("commonroad" in env_id or env_id == "cr-monitor-v0"):
            train_reset_config_path = env_kwargs.pop("train_reset_config_path", PATH_PARAMS["train_reset_config"])
            test_reset_config_path = env_kwargs.pop("test_reset_config_path", PATH_PARAMS["test_reset_config"])
            env = gym.make(env_id,
                           train_reset_config_path=os.path.join(train_reset_config_path, str(rank)),
                           test_reset_config_path=os.path.join(test_reset_config_path, str(rank)),
                           logging_path=logging_path,
                           **env_kwargs)
        else:
            env = gym.make(env_id, logging_path=logging_path, **env_kwargs)

        # We allow a Gym env wrapper (a subclass of gym.Wrapper)
        if wrapper_class:
            env = wrapper_class(env)

        env.seed(seed + rank)
        log_file = os.path.join(log_dir, str(rank)) if log_dir is not None else None
        env = Monitor(env, log_file, info_keywords=info_keywords)

        return env

    return _init


def create_test_env(
        env_id,
        n_envs=1,
        is_atari=False,
        stats_path=None,
        seed=0,
        log_dir="",
        should_render=True,
        hyperparams=None,
        env_kwargs=None,
):
    """
    Create environment for testing a trained agent

    :param env_id: (str)
    :param n_envs: (int) number of processes
    :param is_atari: (bool)
    :param stats_path: (str) path to folder containing saved running averaged
    :param seed: (int) Seed for random number generator
    :param log_dir: (str) Where to log rewards
    :param should_render: (bool) For Pybullet env, display the GUI
    :param env_wrapper: (type) A subclass of gym.Wrapper to wrap the original
                        env with
    :param hyperparams: (dict) Additional hyperparams (ex: n_stack)
    :param env_kwargs: (Dict[str, Any]) Optional keyword argument to pass to the env constructor
    :return: (gym.Env)
    """
    # HACK to save logs
    if log_dir is not None:
        os.environ["OPENAI_LOG_FORMAT"] = "csv"
        os.environ["OPENAI_LOGDIR"] = os.path.abspath(log_dir)
        os.makedirs(log_dir, exist_ok=True)
        logger.configure()

    if hyperparams is None:
        hyperparams = {}

    if env_kwargs is None:
        env_kwargs = {}

    # Create the environment and wrap it if necessary
    env_wrapper = get_wrapper_class(hyperparams)
    if "env_wrapper" in hyperparams.keys():
        del hyperparams["env_wrapper"]

    if is_atari:
        print("Using Atari wrapper")
        env = make_atari_env(env_id, num_env=n_envs, seed=seed)
        # Frame-stacking with 4 frames
        env = VecFrameStack(env, n_stack=4)
    elif n_envs > 1:
        # start_method = 'spawn' for thread safe
        env = SubprocVecEnv(
            [
                make_env(
                    env_id,
                    i,
                    seed,
                    log_dir,
                    wrapper_class=env_wrapper,
                    env_kwargs=env_kwargs,
                )
                for i in range(n_envs)
            ]
        )
    # Pybullet envs does not follow gym.render() interface
    elif "Bullet" in env_id:
        # HACK: force SubprocVecEnv for Bullet env
        env = SubprocVecEnv(
            [
                make_env(
                    env_id,
                    0,
                    seed,
                    log_dir,
                    wrapper_class=env_wrapper,
                    env_kwargs=env_kwargs,
                )
            ]
        )
    else:
        env = DummyVecEnv(
            [
                make_env(
                    env_id,
                    0,
                    seed,
                    log_dir,
                    wrapper_class=env_wrapper,
                    env_kwargs=env_kwargs,
                )
            ]
        )

    # Load saved stats for normalizing input and rewards
    # And optionally stack frames
    if stats_path is not None:
        if hyperparams["normalize"]:
            print("Loading running average")
            print("with params: {}".format(hyperparams["normalize_kwargs"]))
            env = SBVecNormalize(env, training=False, **hyperparams["normalize_kwargs"])

            if os.path.exists(os.path.join(stats_path, "vecnormalize.pkl")):
                env = SBVecNormalize.load(
                    os.path.join(stats_path, "vecnormalize.pkl"), env
                )
                # Deactivate training and reward normalization
                env.training = False
                env.norm_reward = False
            else:
                # Legacy:
                env.load_running_average(stats_path)

        n_stack = hyperparams.get("frame_stack", 0)
        if n_stack > 0:
            print("Stacking {} frames".format(n_stack))
            env = VecFrameStack(env, n_stack)
    return env


def linear_schedule(initial_value):
    """
    Linear learning rate schedule.

    :param initial_value: (float or str)
    :return: (function)
    """
    if isinstance(initial_value, str):
        initial_value = float(initial_value)

    def func(progress):
        """
        Progress will decrease from 1 (beginning) to 0

        :param progress: (float)
        :return: (float)
        """
        return progress * initial_value

    return func


def get_trained_models(log_folder):
    """

    :param log_folder: (str) Root log folder
    :return: (dict) Dict representing the trained agent
    """
    algos = os.listdir(log_folder)
    trained_models = {}
    for algo in algos:
        for ext in ["zip", "pkl"]:
            for env_id in glob.glob("{}/{}/*.{}".format(log_folder, algo, ext)):
                # Retrieve env name
                env_id = env_id.split("/")[-1].split(".{}".format(ext))[0]
                trained_models["{}-{}".format(algo, env_id)] = (algo, env_id)
    return trained_models


def get_latest_run_id(log_path, env_id):
    """
    Returns the latest run number for the given log name and log path,
    by finding the greatest number in the directories.

    :param log_path: (str) path to log folder
    :param env_id: (str)
    :return: (int) latest run number
    """
    max_run_id = 0
    for path in glob.glob(log_path + "/{}_[0-9]*".format(env_id)):
        file_name = path.split("/")[-1]
        ext = file_name.split("_")[-1]
        if (
                env_id == "_".join(file_name.split("_")[:-1])
                and ext.isdigit()
                and int(ext) > max_run_id
        ):
            max_run_id = int(ext)
    return max_run_id


def get_saved_hyperparams(stats_path, norm_reward=False, test_mode=False):
    """

    :param stats_path: (str)
    :param norm_reward: (bool)
    :param test_mode: (bool)
    :return: (dict, str)
    """
    hyperparams = {}
    if not os.path.isdir(stats_path):
        stats_path = None
    else:
        config_file = os.path.join(stats_path, "config.yml")
        if os.path.isfile(config_file):
            # Load saved hyperparameters
            with open(os.path.join(stats_path, "config.yml"), "r") as f:
                hyperparams = yaml.load(
                    f, Loader=yaml.UnsafeLoader
                )  # pytype: disable=module-attr
            hyperparams["normalize"] = hyperparams.get("normalize", False)
        else:
            obs_rms_path = os.path.join(stats_path, "obs_rms.pkl")
            hyperparams["normalize"] = os.path.isfile(obs_rms_path)

        # Load normalization params
        if hyperparams["normalize"]:
            if isinstance(hyperparams["normalize"], str):
                normalize_kwargs = eval(hyperparams["normalize"])
                if test_mode:
                    normalize_kwargs["norm_reward"] = norm_reward
            else:
                normalize_kwargs = {
                    "norm_obs": hyperparams["normalize"],
                    "norm_reward": norm_reward,
                }
            hyperparams["normalize_kwargs"] = normalize_kwargs
    return hyperparams, stats_path


def find_saved_model(algo, log_path, env_id, load_best=False):
    """

    :param algo: (str)
    :param log_path: (str) Path to the directory with the saved model
    :param env_id: (str)
    :param load_best: (bool)
    :return: (str) Path to the saved model
    """
    model_path, found = None, False
    for ext in ["pkl", "zip"]:
        model_path = "{}/{}.{}".format(log_path, env_id, ext)
        found = os.path.isfile(model_path)
        if found:
            break

    if load_best:
        model_path = os.path.join(log_path, "best_model.zip")
        found = os.path.isfile(model_path)

    if not found:
        raise ValueError(
            "No model found for {} on {}, path: {}".format(algo, env_id, model_path)
        )
    return model_path


class StoreDict(argparse.Action):
    """
    Custom argparse action for storing dict.

    In: args1:0.0 args2:"dict(a=1)"
    Out: {'args1': 0.0, arg2: dict(a=1)}
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreDict, self).__init__(option_strings, dest, nargs=nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        arg_dict = {}
        for arguments in values:
            key = arguments.split(":")[0]
            value = ":".join(arguments.split(":")[1:])
            # Evaluate the string as python code
            arg_dict[key] = eval(value)
            setattr(namespace, self.dest, arg_dict)

if __name__ == "__main__":
    with open("model_hyperparameters.yml", "r") as f:
        hyperparams = yaml.load(f, Loader=yaml.Loader)
    get_wrapper_class(hyperparams)




"""
Unit tests of the module gym_commonroad.observation
"""

import numpy as np
from commonroad.scenario.scenario import Scenario, ScenarioID
from commonroad.scenario.trajectory import CustomState
from commonroad_dc import pycrcc
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object

from commonroad_rl.tests.common.marker import *

# ====================================================================================================================
#
#                                    ObservationCollector
#
# ====================================================================================================================

from typing import List, Tuple
from commonroad.scenario.lanelet import LaneletNetwork, Lanelet
from commonroad_rl.gym_commonroad.observation import ObservationCollector
from commonroad_rl.gym_commonroad.utils.navigator import Navigator


@pytest.mark.parametrize(("orientation", "lanelet_vertices", "expected_id"), [
    (6.280, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),  # orientation 0.0... uh oh!
             ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),  # orientation 1.57 should not be closer!
             ], 1,), (3 * np.pi / 2, [([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),
                                      # orientation -1.57 should be closest
                                      ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),
                                      # orientation 1.57 should not be closer!
                                      ], 1,),  # some trivial cases
    (0, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]])], 1),
    (0, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),  # orientation 0 a perfect match!
         ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),  # orientation 1.57, way too far away!
         ], 1,), (0, [([[-1, 1], [1, 1]], [[-1, 0], [1, -0.01]], [[-1, -1], [1, -1]],),
                      # orientation 3.14... uh oh! => handled to -0.01
                      ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),
                      # orientation 1.57 should not be closer!
                      ], 1,),
    (0, [([[-1, 1], [1, 1]], [[-1, 0], [1, -0.01]], [[-1, -1], [1, -1]],),  # orientation 0.01... uh oh!
         ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),  # orientation 1.57 should not be closer!
         ], 1,),
    (1.570, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),  # orientation 0 should not be closer
             ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],),  # orientation 1.57 should  be perfect
             ], 2,),
    (1.570, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),  # orientation 0 should not be closer
             ([[-1, -1], [-1, 1]], [[0, -1], [0.01, 1]], [[1, -1], [1, 1]],),  # orientation 1.57 should  be perfect
             ], 2,),
    (1.570, [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),  # orientation 0 should not be closer
             ([[-1, -1], [-1, 1]], [[0, -1], [-0.01, 1]], [[1, -1], [1, 1]],),  # orientation 1.57 should  be perfect
             ], 2,), ], )
@unit_test
@functional
def test_get_lanelet_id_by_state_orientation(orientation: float, lanelet_vertices: List[Tuple[List[List[float]]]],
                                             expected_id: int, ):
    """
    Test that get lanelet id by state makes correct use of orientations
    """
    s = CustomState(position=np.array([0, 0]), orientation=orientation, time_step=0)
    scenario = Scenario(0.1, ScenarioID("test_id"))
    for i, lv in enumerate(lanelet_vertices, start=1):
        l = Lanelet(np.array(lv[0]), np.array(lv[1]), np.array(lv[2]), i)
        scenario.add_objects(l)

    lanelet_polygons = [(l.lanelet_id, l.convert_to_polygon()) for l in scenario.lanelet_network.lanelets]
    lanelet_polygons_sg = pycrcc.ShapeGroup()
    for l_id, poly in lanelet_polygons:
        lanelet_polygons_sg.add_shape(create_collision_object(poly))
    lanelet_ids = Navigator.sorted_lanelet_ids(
        ObservationCollector._related_lanelets_by_state(s, lanelet_polygons, lanelet_polygons_sg),
        s.orientation, s.position, scenario)
    lanelet_id = -1 if len(lanelet_ids) == 0 else lanelet_ids[0]
    assert expected_id == lanelet_id


@pytest.mark.parametrize(("lanelet_vertices", "expected_ids"), [(
        [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),
         ([[-1, -1], [-1, 1]], [[0, -1], [0, 1]], [[1, -1], [1, 1]],), ], [1, 2],), (
        [([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],),
         ([[2, 2], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],), ], [1],), (
        [([[2, 2], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],),
         ([[2, 2], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],), ], [],), (
        [([[2, 2], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],),
         ([[2, 2], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],),
         ([[2, 5], [2, 4]], [[3, 2], [3, 4]], [[4, 2], [4, 4]],),
         ([[-1, 1], [1, 1]], [[-1, 0], [1, 0]], [[-1, -1], [1, -1]],), ], [4],), ])
@unit_test
@functional
def test_related_lanelets_by_state(lanelet_vertices: List[Tuple[List[List[float]]]], expected_ids: List[int], ):
    """
    Test that related lanelet id by state
    """
    s = CustomState(position=np.array([0, 0]), orientation=0, time_step=0)
    scenario = Scenario(dt=0.1, scenario_id=ScenarioID("test_id"))
    for i, lv in enumerate(lanelet_vertices, start=1):
        l = Lanelet(np.array(lv[0]), np.array(lv[1]), np.array(lv[2]), i)
        scenario.add_objects(l)

    lanelet_polygons = [(l.lanelet_id, l.convert_to_polygon()) for l in scenario.lanelet_network.lanelets]
    lanelet_polygons_sg = pycrcc.ShapeGroup()
    for l_id, poly in lanelet_polygons:
        lanelet_polygons_sg.add_shape(create_collision_object(poly))

    res = ObservationCollector._related_lanelets_by_state(s, lanelet_polygons, lanelet_polygons_sg)
    assert isinstance(res, list)
    assert all([a == b for a, b in zip(res, expected_ids)])


# TODO: update expected value after ccosy update
@pytest.mark.parametrize(("lanelets",
                          "cartesian_coordinates", "expected_curvilinear_coordinates", "desired_merged_lanelet"),
                         [([Lanelet(lanelet_id=0, left_vertices=np.array([[-1, 1], [1, 1]]),
                                    center_vertices=np.array([[-1, 0], [1, 0]]),
                                    right_vertices=np.array([[-1, -1], [1, -1]]), adjacent_right_same_direction=1),
                            Lanelet(lanelet_id=1, left_vertices=np.array([[-1, -1], [1, -1]]),
                                    center_vertices=np.array([[-1, -2], [1, -2]]),
                                    right_vertices=np.array([[-1, -3], [1, -3]]), adjacent_left_same_direction=0)],
                           np.array([0, 0]), np.array([1, 2]), 1),
                          ([Lanelet(lanelet_id=0, left_vertices=np.array([[-1, 1], [1, 1]]),
                                    center_vertices=np.array([[-1, 0], [1, 0]]),
                                    right_vertices=np.array([[-1, -1], [1, -1]]),
                                    successor=[1]),
                            Lanelet(lanelet_id=1, left_vertices=np.array([[1, 1], [3, 1]]),
                                    center_vertices=np.array([[1, 0], [3, 0]]),
                                    right_vertices=np.array([[1, -1], [3, -1]]),
                                    predecessor=[0])],
                           np.array([0, 0]), np.array([1, 0]), 1),
                          ([Lanelet(lanelet_id=0, left_vertices=np.array([[-1, 1], [1, 1]]),
                                    center_vertices=np.array([[-1, 0], [1, 0]]),
                                    right_vertices=np.array([[-1, -1], [1, -1]]), successor=[1],
                                    adjacent_right_same_direction=2),
                            Lanelet(lanelet_id=1, left_vertices=np.array([[1, 1], [3, 1]]),
                                    center_vertices=np.array([[1, 0], [3, 0]]),
                                    right_vertices=np.array([[1, -1], [3, -1]]), predecessor=[0]),
                            Lanelet(lanelet_id=2, left_vertices=np.array([[-1, -1], [1, -1]]),
                                    center_vertices=np.array([[-1, -2], [1, -2]]),
                                    right_vertices=np.array([[-1, -3], [1, -3]]), adjacent_left_same_direction=0,
                                    adjacent_right_same_direction=3),
                            Lanelet(lanelet_id=3, left_vertices=np.array([[-1, -3], [1, -3]]),
                                    center_vertices=np.array([[-1, -4], [1, -4]]),
                                    right_vertices=np.array([[-1, -5], [1, -5]]), adjacent_left_same_direction=2)],
                           np.array([2, 4]), np.array([3, 4]), 1)])
@unit_test
@functional
def test_get_local_curvi_cosy(lanelets: List[Lanelet], cartesian_coordinates: np.ndarray,
                              expected_curvilinear_coordinates: np.ndarray, desired_merged_lanelet: int):
    """test ObservationCollector get_local_curvi_cosy function"""
    ego_vehicle_lanelet_id = 1
    max_lane_merge_range = 5000.

    lanelet_network = LaneletNetwork()
    for lanelet in lanelets:
        lanelet_network.add_lanelet(lanelet)

    ref_path_dict = dict()

    curvi_cosy, ref_merged_lanelet = ObservationCollector.get_local_curvi_cosy(lanelet_network, ego_vehicle_lanelet_id,
                                                                               ref_path_dict, max_lane_merge_range)
    calculated_curv_coords = curvi_cosy.convert_to_curvilinear_coords(*cartesian_coordinates)

    assert all(np.isclose(calculated_curv_coords, expected_curvilinear_coordinates, atol=1e-3))
    if not ref_merged_lanelet is None:
        assert np.isclose(ref_merged_lanelet.lanelet_id, desired_merged_lanelet)
    else:
        assert desired_merged_lanelet is None



import copy
from collections import defaultdict

from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario
from commonroad_dc.boundary.boundary import create_road_boundary_obstacle

from commonroad_rl.gym_commonroad.utils.scenario import get_road_edge


def get_all_connected_lanelets(scenario: Scenario) -> dict:
    """
    Create all possible lanes by merging predecessors and successors, then create a dict with its keys as lanelet id
    and values as connected lanelet ids.
    
    :return: dict
    """
    merged_lanelet_dict = defaultdict(set)
    for l in scenario.lanelet_network.lanelets:  # iterate in all lanelet in this scenario
        if not l.predecessor and not l.successor:  # the lanelet is a lane itself
            merged_lanelet_dict[l.lanelet_id].add(l.lanelet_id)
        elif not l.predecessor:
            max_lane_merge_range = 1000.0
            _, sub_lanelet_ids = Lanelet.all_lanelets_by_merging_successors_from_lanelet(l, scenario.lanelet_network,
                                                                                         max_lane_merge_range)
            for s in sub_lanelet_ids:
                for i in s:
                    merged_lanelet_dict[i].update(s)
    return merged_lanelet_dict


def generate_reset_config(scenario: Scenario, open_lane_ends) -> dict:
    """
    Generate a dict of reset configurations which contains obstacle lanelet ids, road edge, collision checker,
    lanelet boundary and lanelet connection dict.

    :param scenario: commonroad scenario
    :return:
    """
    (left_road_edge_lanelet_id, left_road_edge, right_road_edge_lanelet_id, right_road_edge) = get_road_edge(scenario)
    _, lanelet_boundary = create_road_boundary_obstacle(scenario, method="obb_rectangles", open_lane_ends=open_lane_ends)
    connected_lanelet_dict = get_all_connected_lanelets(scenario)

    meta_scenario = copy.deepcopy(scenario)
    meta_scenario.remove_obstacle(scenario.obstacles)
    reset_config = {"left_road_edge_lanelet_id_dict": left_road_edge_lanelet_id,
                    "left_road_edge_dict": left_road_edge,
                    "right_road_edge_lanelet_id_dict": right_road_edge_lanelet_id,
                    "right_road_edge_dict": right_road_edge,
                    "boundary_collision_object": lanelet_boundary,
                    "connected_lanelet_dict": connected_lanelet_dict,
                    "meta_scenario": meta_scenario}

    return reset_config



import os
import glob
import shutil
import pickle
import argparse
import multiprocessing as mp

from time import time
from typing import List

from commonroad.common.file_reader import CommonRoadFileReader
import commonroad_dc.pycrcc as pycrcc
from commonroad_rl.tools.pickle_scenario.preprocessing import generate_reset_config
from commonroad_rl.gym_commonroad.utils.scenario import parse_map_name


def get_args():
    parser = argparse.ArgumentParser(description="Converts CommonRoad xml files to pickle files",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--input_dir", "-i", type=str, default="/data/highD-dataset-v1.0/cr_scenarios")
    parser.add_argument("--output_dir", "-o", type=str, default="/data/highD-dataset-v1.0/pickles")
    parser.add_argument("--duplicate", "-d", action="store_true",
                        help="Duplicate one scenario file to problem_train and problem_test, for overfitting")
    parser.add_argument('--num_processes', type=int, default=1,
                        help='Number of multiple processes to convert dataset, default=1')

    return parser.parse_args()


def is_initial_collision(planning_problem_set, env_reset):

    initial_state = list(planning_problem_set.planning_problem_dict.values())[0].initial_state
    boundary_collision_object = env_reset["boundary_collision_object"]
    ego_collision_object = pycrcc.RectOBB(
        4.508 / 2,
        1.610 / 2,
        initial_state.orientation if hasattr(initial_state, "orientation") else 0.0,
        initial_state.position[0],
        initial_state.position[1],
        )

    return ego_collision_object.collide(boundary_collision_object)


def process_single_file(rank: int, fns: List[str], duplicate: bool, output_dir: str, open_lane_ends: bool):

    meta_scenario_reset_dict = dict()
    processed_location_list = []

    for i, fn in enumerate(fns):
        print(f"{i + 1}/{len(fns)}", end="\r")
        scenario, planning_problem_set = CommonRoadFileReader(fn).open(lanelet_assignment=True)
        problem_dict = {"obstacle": scenario.obstacles, "planning_problem_set": planning_problem_set}
        map_name_id = parse_map_name(scenario.scenario_id)
        if map_name_id not in processed_location_list:
            processed_location_list.append(map_name_id)
            env_reset = generate_reset_config(scenario, open_lane_ends)
            if is_initial_collision(planning_problem_set, env_reset):
                continue
            meta_scenario_reset_dict[map_name_id] = env_reset

        if duplicate:
            with open(os.path.join(output_dir, "problem_train", f"{scenario.scenario_id}.pickle"), "wb") as f:
                pickle.dump(problem_dict, f)
            shutil.copytree(os.path.join(output_dir, "problem_train"), os.path.join(output_dir, "problem_test"))
        else:
            with open(os.path.join(output_dir, "problem", f"{scenario.scenario_id}.pickle"), "wb") as f:
                pickle.dump(problem_dict, f)

    os.makedirs(os.path.join(output_dir, f"meta_scenario_{rank}"), exist_ok=True)
    with open(os.path.join(output_dir, f"meta_scenario_{rank}", "meta_scenario_reset_dict.pickle"), "wb") as f:
        pickle.dump(meta_scenario_reset_dict, f)


def pickle_xml_scenarios(input_dir: str, output_dir: str, duplicate: bool = False, num_processes: int = 1,
                         open_lane_ends: bool = True):
    # makedir
    os.makedirs(output_dir, exist_ok=True)

    meta_scenario_path = "meta_scenario"
    fns = glob.glob(os.path.join(input_dir, "*.xml"))

    os.makedirs(os.path.join(output_dir, meta_scenario_path), exist_ok=True)
    if not duplicate:
        os.makedirs(os.path.join(output_dir, "problem"), exist_ok=True)
    else:
        os.makedirs(os.path.join(output_dir, "problem_train"), exist_ok=True)

    _start_time = time()
    num_scenarios_per_process = len(fns) // num_processes

    if num_processes > 1:
        with mp.Pool(processes=num_processes) as pool:
            pool.starmap(
                process_single_file,
                [
                    (
                        i,
                        fns[i * num_scenarios_per_process : (i + 1) * num_scenarios_per_process],
                        duplicate,
                        output_dir, open_lane_ends
                    )
                    for i in range(num_processes)
                ]
            )
    else:
        process_single_file(0, fns, duplicate, output_dir, open_lane_ends)
    meta_scenario_reset_dict = {}
    for i in range(num_processes):
        with open(os.path.join(output_dir, f"{meta_scenario_path}_{i}", "meta_scenario_reset_dict.pickle"), "rb") as f:
            meta_scenario_reset_dict.update(pickle.load(f))
        shutil.rmtree(os.path.join(output_dir, f"{meta_scenario_path}_{i}"))

    with open(os.path.join(output_dir, meta_scenario_path, "meta_scenario_reset_dict.pickle"), "wb") as f:
        pickle.dump(meta_scenario_reset_dict, f)

    print(len(meta_scenario_reset_dict.keys()))
    print("Took {}s".format(time() - _start_time))


if __name__ == "__main__":
    # get arguments
    args = get_args()

    pickle_xml_scenarios(
        args.input_dir,
        args.output_dir,
        duplicate=args.duplicate,
        num_processes=args.num_processes,
        open_lane_ends=True
    )



import warnings
from collections import OrderedDict
from typing import Union, Dict, List, Tuple

import gym
import numpy as np
from commonroad.geometry.shape import ShapeGroup, Shape
from commonroad.planning.goal import GoalRegion
from commonroad.planning.planning_problem import PlanningProblem
from commonroad.scenario.obstacle import State
from commonroad.scenario.scenario import Scenario
from commonroad.visualization.mp_renderer import MPRenderer
from commonroad.visualization.draw_params import LaneletParams
from commonroad_rl.gym_commonroad.observation.observation import Observation
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from commonroad_dc.pycrccosy import CurvilinearCoordinateSystem

class GoalObservation(Observation):
    """
    This class contains all helper methods and the main observation method for observations related to the goal

    :param configs: the configuration specification
    """

    def __init__(self, configs: Dict, config_name: str = "goal_configs"):
        # Read configs
        configs = configs[config_name]
        self.relax_is_goal_reached: bool = configs.get("relax_is_goal_reached")

        self.observe_distance_goal_long: bool = configs.get("observe_distance_goal_long")
        self.observe_distance_goal_lat: bool = configs.get("observe_distance_goal_lat")
        self.observe_distance_goal_long_lane: bool = configs.get("observe_distance_goal_long_lane")
        self.observe_distance_goal_time: bool = configs.get("observe_distance_goal_time")
        self.observe_distance_goal_orientation: bool = configs.get("observe_distance_goal_orientation")
        self.observe_distance_goal_velocity: bool = configs.get("observe_distance_goal_velocity")
        self.observe_euclidean_distance: bool = configs.get("observe_euclidean_distance")

        self.observe_is_time_out = configs.get("observe_is_time_out")
        self.observe_is_goal_reached = configs.get("observe_is_goal_reached")

        # location for storing the past observations
        self.observation_history_dict: dict = dict()

    def build_observation_space(self) -> OrderedDict:
        observation_space_dict = OrderedDict()

        if self.observe_euclidean_distance:
            observation_space_dict["euclidean_distance"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_distance_goal_long:
            observation_space_dict["distance_goal_long"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
            observation_space_dict["distance_goal_long_advance"] = gym.spaces.Box(-np.inf, np.inf, (1,),
                                                                                  dtype=np.float32)
        if self.observe_distance_goal_lat:
            observation_space_dict["distance_goal_lat"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
            observation_space_dict["distance_goal_lat_advance"] = gym.spaces.Box(-np.inf, np.inf, (1,),
                                                                                 dtype=np.float32)
        if self.observe_distance_goal_long_lane:
            observation_space_dict["distance_goal_long_lane"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_distance_goal_time:
            observation_space_dict["distance_goal_time"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_distance_goal_orientation:
            observation_space_dict["distance_goal_orientation"] = gym.spaces.Box(-np.inf, np.inf, (1,),
                                                                                 dtype=np.float32)
        if self.observe_distance_goal_velocity:
            observation_space_dict["distance_goal_velocity"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_is_goal_reached:
            observation_space_dict["is_goal_reached"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)

        if self.observe_is_time_out:
            observation_space_dict["is_time_out"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)

        return observation_space_dict

    def observe(self, ego_state: State, goal: Union[GoalRegion, None] = None,
                scenario: Union[Scenario, None] = None, planning_problem: Union[PlanningProblem, None] = None,
                ego_lanelet_ids: Union[List[int], None] = None, navigator: Union[Navigator, None] = None,
                episode_length: int = None,local_ccosy: Union[None, CurvilinearCoordinateSystem] = None) \
            -> Union[np.array, Dict]:
        """ Create goal related observation for given state in an environment.

            :param ego_state: state from which to observe the environment
            :param goal: goal region
            :param scenario: current Scenario
            :param planning_problem: the current planning problem
            :return: ndarray of observation if flatten == True, observation dict otherwise
        """
        observation_dict = {}

        if self.observe_euclidean_distance:
            distance = GoalObservation._get_goal_euclidean_distance(ego_state.position, goal)
            observation_dict['euclidean_distance'] = np.array([distance])
        # observe distance goal long and lat and also the advance versions
        if self.observe_distance_goal_long or self.observe_distance_goal_lat:
            distance_goal_long, distance_goal_lat = self.get_long_lat_distance_to_goal(ego_state.position, navigator)

            observation_dict["distance_goal_long"] = np.array([distance_goal_long])
            observation_dict["distance_goal_lat"] = np.array([distance_goal_lat])

            if distance_goal_long is np.nan:
                distance_goal_long = self.observation_history_dict.get("distance_goal_long", 1e4)
                distance_goal_lat = self.observation_history_dict.get("distance_goal_lat", 1e4)
            (distance_goal_long_advance, distance_goal_lat_advance) = \
                self._get_long_lat_distance_advance_to_goal(distance_goal_long, distance_goal_lat)

            self.observation_history_dict["distance_goal_long_advance"] = distance_goal_long_advance
            self.observation_history_dict["distance_goal_lat_advance"] = distance_goal_lat_advance
            observation_dict["distance_goal_long_advance"] = np.array([distance_goal_long_advance])
            observation_dict["distance_goal_lat_advance"] = np.array([distance_goal_lat_advance])

            self.observation_history_dict["distance_goal_long"] = distance_goal_long
            self.observation_history_dict["distance_goal_lat"] = distance_goal_lat

        # observe the time to the goal state
        if self.observe_distance_goal_time:
            distance_goal_time = GoalObservation._get_goal_time_distance(ego_state.time_step, goal)
            observation_dict["distance_goal_time"] = np.array([distance_goal_time])
        # observe distance to the goal orientation from ego
        if self.observe_distance_goal_orientation:
            ego_state_orientation = ego_state.orientation if hasattr(ego_state, "orientation") else \
                np.arctan2(ego_state.velocity_y, ego_state.velocity)
            distance_goal_orientation = GoalObservation._get_goal_orientation_distance(ego_state_orientation, goal)
            observation_dict["distance_goal_orientation"] = np.array([distance_goal_orientation])
        # observe difference between ego velocity and goal velocity
        if self.observe_distance_goal_velocity:
            distance_goal_velocity = GoalObservation._get_goal_velocity_distance(ego_state.velocity, goal)
            observation_dict["distance_goal_velocity"] = np.array([distance_goal_velocity])

        # Get the longitudinal distance until the lane change must be finished
        if self.observe_distance_goal_long_lane:
            distance_goal_long_lane = self._get_long_distance_until_lane_change(ego_state, ego_lanelet_ids,
                                                                                navigator)
            self.observation_history_dict["distance_goal_long_lane"] = distance_goal_long_lane
            observation_dict["distance_goal_long_lane"] = np.array([distance_goal_long_lane])

        # Check if the ego vehicle has reached the goal
        if self.observe_is_goal_reached or self.observe_is_time_out:
            is_goal_reached = self._check_goal_reached(goal, ego_state, self.relax_is_goal_reached)
        if self.observe_is_goal_reached:
            observation_dict["is_goal_reached"] = np.array([is_goal_reached])
        # Check if maximum episode length exceeded
        if self.observe_is_time_out:
            is_time_out = GoalObservation._check_is_time_out(ego_state, goal, is_goal_reached, episode_length)
            if not is_time_out:
                # check if ego vehicle reaches the end of the road
                if not local_ccosy.cartesian_point_inside_projection_domain(ego_state.position[0], ego_state.position[1]):
                    is_time_out=True
            observation_dict["is_time_out"] = np.array([is_time_out])

        return observation_dict

    def draw(self, render_configs: Dict, render: MPRenderer, navigator: Union[None, Navigator]):
        """ Method to draw the observation """
        if render_configs["render_global_ccosy"]:
            # TODO: This functionality has been taken from commonroad-route-planner
            # As soon as the route-planner supports drawing only the ccosy, this part should be replaced
            draw_params = LaneletParams()
            draw_params.center_bound_color = "#128c01"
            draw_params.draw_line_markings = False
            draw_params.draw_left_bound = False
            draw_params.draw_right_bound = False
            draw_params.draw_start_and_direction = False
            draw_params.draw_linewidth = 1.
            draw_params.fill_lanelet = False
            draw_params.facecolor = "#128c01"

            # TODO: temporary fix for missing draw method in Lanelet, remove after fixed in cr-io
            from commonroad.scenario.lanelet import LaneletNetwork
            LaneletNetwork.create_from_lanelet_list(navigator.merged_route_lanelets).draw(render, draw_params=draw_params)
            # for route_merged_lanelet in navigator.merged_route_lanelets:
            #     route_merged_lanelet.draw(render, draw_params=draw_params)

    @staticmethod
    def _get_goal_euclidean_distance(position: np.array, goal: GoalRegion) -> float:
        """
        calculates the euclidean distance of the current position to the goal

        :param position: current position
        :param goal: the goal of the current planning problem
        :return euclidean distance
        """
        if "position" not in goal.state_list[0].attributes:
            return 0.

        else:
            f_pos = goal.state_list[0].position
            if isinstance(f_pos, ShapeGroup):
                goal_position_list = np.array(
                    [GoalObservation._convert_shape_group_to_center(s.position) for s in goal.state_list])
            elif isinstance(f_pos, Shape):
                goal_position_list = np.array([s.position.center for s in goal.state_list])
            else:
                warnings.warn(f"Trying to calculate relative goal orientation but goal state position "
                              f"type ({type(f_pos)}) is not support, please set "
                              f"observe_distance_goal_euclidean = False or "
                              f"change state position type to one of the following: Polygon, Rectangle, Circle")
                return 0.
            goal_position_mean = np.mean(goal_position_list, axis=0)
            return np.linalg.norm(position - goal_position_mean)

    @staticmethod
    def _convert_shape_group_to_center(shape_group: ShapeGroup):
        position_list = [shape.center for shape in shape_group.shapes]
        return np.mean(np.array(position_list), axis=0)

    def _get_long_lat_distance_advance_to_goal(self, distance_goal_long: float,
                                               distance_goal_lat: float) -> Tuple[float, float]:
        """
        Get longitudinal and lateral distances to the goal over the planned route

        :param distance_goal_long: the current distance_goal_long observation
        :param distance_goal_lat: the current distance_goal_lat observation

        :return: The tuple of the longitudinal and the lateral distance advances
        """
        if "distance_goal_long" not in self.observation_history_dict or not self.observe_distance_goal_long:
            distance_goal_long_advance = 0.0
        else:
            distance_goal_long_advance = abs(self.observation_history_dict["distance_goal_long"]) - abs(
                distance_goal_long)

        if "distance_goal_lat" not in self.observation_history_dict or not self.observe_distance_goal_lat:
            distance_goal_lat_advance = 0.0
        else:
            distance_goal_lat_advance = abs(self.observation_history_dict["distance_goal_lat"]) - abs(distance_goal_lat)

        return distance_goal_long_advance, distance_goal_lat_advance

    @staticmethod
    def _get_goal_velocity_distance(velocity: float, goal: GoalRegion) -> float:
        """
        calculates the difference to the goal velocity

        calculates velocity - goal_velocity_interval_start    if velocity < goal_velocity_interval_start
                   velocity - goal_velocity_interval_end    if velocity > goal_velocity_interval_end
                   0                                          else

        :param velocity: velocity of current state
        :param goal: GoalRegion of current planning problem
        :return difference to the nearest goal velocity boundary
        """
        if "velocity" not in goal.state_list[0].attributes:
            return 0.

        else:
            velocity_start_list = np.array([s.velocity.start for s in goal.state_list])
            velocity_end_list = np.array([s.velocity.end for s in goal.state_list])
            goal_velocity_interval_start: float = np.squeeze(np.mean(velocity_start_list))
            goal_velocity_interval_end: float = np.squeeze(np.mean(velocity_end_list))
            if velocity < goal_velocity_interval_start:
                return velocity - goal_velocity_interval_start
            elif velocity > goal_velocity_interval_end:
                return velocity - goal_velocity_interval_end
            else:
                return 0.

    @staticmethod
    def _get_goal_orientation_distance(orientation: float, goal: GoalRegion) -> float:
        """
        calculate the distance of the current vehicle orientation to the goal

        calculates orientation - goal_orientation_interval_start    if orientation < goal_orientation_interval_start
                   orientation - goal_orientation_interval_end      if orientation > goal_orientation_interval_start
                   0                                                else

        :param orientation: orientation of current state
        :param goal: GoalRegion of current planning problem
        :return difference to the nearest goal orientation boundary using radians in interval [-pi,pi]
        """
        if "orientation" not in goal.state_list[0].attributes:
            return 0.

        else:
            orientation_start_list = np.array([s.orientation.start for s in goal.state_list])
            orientation_end_list = np.array([s.orientation.end for s in goal.state_list])
            goal_orientation_interval_start: float = np.mean(orientation_start_list) % (2 * np.pi)
            goal_orientation_interval_end: float = np.mean(orientation_end_list) % (2 * np.pi)

            orientation = orientation % (2 * np.pi)
            distance_start_right: float = (orientation - goal_orientation_interval_start) % (2 * np.pi)
            distance_start_left: float = (2 * np.pi) - distance_start_right
            distance_end_left: float = (goal_orientation_interval_end - orientation) % (2 * np.pi)
            distance_end_right: float = (2 * np.pi) - distance_end_left

            if (distance_end_left + distance_start_right) <= 2 * np.pi:
                return 0.0
            elif distance_start_left <= distance_end_right:
                return distance_start_left
            else:
                return -distance_end_right

    def _get_long_distance_until_lane_change(self, ego_state: State, ego_vehicle_lanelet_ids: List[int],
                                             navigator: Navigator) -> float:
        """
        Get the longitudinal distance until the lane change must be finished. It means that the ego vehicle is
        allowed to continue its was in the current lanelet and in its adjacent, but after this returned value,
        it must change the lane to the one which successor will lead to the goal

        :param ego_state: The current state
        :param ego_vehicle_lanelet_ids: The lanelet ids of the current state
        :param navigator: the navigator of the current planning problem
        :raises AssertionError: if ValueError of goal.navigator.get_lane_change_distance and no observations stored in
                history -> Ego vehicle started outside the global coordinate system
        :return: The longitudinal distance until the lane change must be finished
        """
        try:
            return navigator.get_lane_change_distance(ego_state, ego_vehicle_lanelet_ids)
            if not hasattr(ego_state, "orientation"):
                setattr(ego_state, "orientation", np.arctan2(ego_state.velocity_y, ego_state.velocity))
        except ValueError:
            assert self.distance_goal_long_lane, \
                "Ego vehicle started outside the global coordinate system"
            return self.observation_history_dict["distance_goal_long_lane"]

    @staticmethod
    def _get_goal_time_distance(time_step: float, goal: GoalRegion) -> float:
        """
        calculates the remaining time till the start of the goal time interval

        calculates time_step - goal_time_step_interval_start    if time_step < goal_time_step_interval_start
                   time_step - goal_time_interval_end           if time_step > goal_time_interval_end
                   0                                            else

        :param time_step: current time step
        :param goal: GoalRegion of current planning problem
        :return difference to the nearest goal time boundary
        """
        if "time_step" not in goal.state_list[0].attributes:
            return 0

        # time_step is mandatory for GoalRegion, doesn't need to check attribute
        time_start_list = np.array([s.time_step.start for s in goal.state_list])
        time_end_list = np.array([s.time_step.end for s in goal.state_list])
        goal_time_interval_start: float = np.mean(time_start_list)
        goal_time_interval_end: float = np.mean(time_end_list)
        if time_step < goal_time_interval_start:
            return time_step - goal_time_interval_start
        elif time_step > goal_time_interval_end:
            return time_step - goal_time_interval_end
        else:
            return 0

    @staticmethod
    def _check_goal_reached(goal: GoalRegion, ego_state: State, relax_is_goal_reached: bool = False) -> bool:
        """
        Check if goal is reached by ego vehicle.

        :param goal: GoalRegion of current planning problem
        :param ego_state: state of ego vehicle.
        :param relax_is_goal_reached: relaxes goal specification to position goal only
        :return: True if goal is reached
        """
        if ego_state.time_step == 0:
            return False

        if relax_is_goal_reached:
            for state in goal.state_list:
                if state.position.contains_point(ego_state.position):
                    return True
            return False
        else:
            return goal.is_reached(ego_state)

    @staticmethod
    def _check_is_time_out(ego_state: State, goal: GoalRegion, is_goal_reached: bool, episode_length=None) -> bool:
        """
        Check if maximum episode length exceeded

        :param goal: GoalRegion of current planning problem
        :param ego_state: state of ego vehicle
        :param is_goal_reached: flag whether the goal has been reached already
        :return: True if no more time left and the goal has not been reached
        """
        # TODO maybe store value and only calculate once
        if episode_length is None:
            episode_length = max(s.time_step.end for s in goal.state_list)
        return ego_state.time_step >= episode_length and not is_goal_reached

    @staticmethod
    def get_long_lat_distance_to_goal(position: np.array, navigator: Navigator) -> Tuple[float, float]:
        """
        Get longitudinal and lateral distances to the goal over the planned route

        :param position: the current position of the agent
        :param navigator: the navigator of the current planning problem
        :raises AssertionError: an AssertionError will be raised if goal.navigator.get_lon_lat_distance_to_goal
                raises a ValueError and there have not been any observations stored in the history

                this means that the ego vehicle started outside the global coordinate system

        :return: The tuple of the longitudinal and the lateral distances
        """
        try:
            return navigator.get_long_lat_distance_to_goal(position)
        except ValueError:
            return np.nan, np.nan


if __name__ == "__main__":
    import yaml
    from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

    config_file = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
        config = yaml.safe_load(config_file)
    configs = config["env_configs"]
    goal_observation = GoalObservation(configs)
    print(goal_observation)



"""Unit tests for hybrid reward class"""
import math
import os
import pickle

import numpy as np
import yaml
from commonroad.common.solution import VehicleModel
from commonroad.scenario.scenario import ScenarioID

from commonroad_rl.gym_commonroad.reward.hybrid_reward import HybridReward
from commonroad_rl.gym_commonroad.utils.scenario import parse_map_name
from commonroad_rl.gym_commonroad.action import DiscretePMJerkAction
from commonroad_rl.gym_commonroad import constants
from commonroad_rl.gym_commonroad.observation import ObservationCollector
from commonroad_rl.gym_commonroad.utils.scenario_io import restore_scenario
from commonroad_rl.tests.common.path import resource_root, output_root
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios

configs = {
    "reward_type": "hybrid_reward",
    "reward_configs": {"hybrid_reward":
        {
            "reward_goal_reached": 50.,
            "reward_collision": -50.,
            "reward_off_road": -20.,
            "reward_friction_violation": -30.,
            "reward_time_out": -10.,
            "reward_closer_to_goal_long": 0.5,
            "reward_closer_to_goal_lat": 2.5,
            "reward_get_close_goal_time": 0.5,
            "reward_close_goal_orientation": 0.5,
            "reward_close_goal_velocity": 0.5,
            "reward_orientation_to_ref": 1.
        }
    },
    "goal_configs": {
        "observe_euclidean_distance": False,
        "observe_distance_goal_long": False
    },
    "surrounding_configs": {
        "observe_lane_change": False,
        "observe_lidar_circle_surrounding": True,
        "lidar_sensor_radius": 50.
    },
    "lanelet_configs": {
        "observe_route_reference_path": True,
        "distances_route_reference_path": [-1000, 0, 5, 15, 100]
    }
}


@unit_test
@functional
def test_termination_reward():
    reward = HybridReward(configs)
    observations = {
        "is_goal_reached": [True],
        "is_collision": [False],
        "is_off_road": [False],
        "is_friction_violation": [False],
        "is_time_out": [False]
    }

    assert reward.termination_reward(observations) == 50

    observations = {
        "is_goal_reached": [False],
        "is_collision": [False],
        "is_off_road": [True],
        "is_friction_violation": [False],
        "is_time_out": [False]
    }

    assert reward.termination_reward(observations) == -20


@unit_test
@functional
def test_goal_distance_reward():
    reward = HybridReward(configs)
    observations = {
        "distance_goal_long_advance": [3.],
        "distance_goal_lat_advance": [1.]
    }

    assert math.isclose(reward.goal_distance_reward(observations), 4.)

    observations = {
        "distance_goal_long_advance": [0.5],
        "distance_goal_lat_advance": [0.]
    }

    assert math.isclose(reward.goal_distance_reward(observations), .25)


@unit_test
@functional
def test_goal_time_reward():
    reward = HybridReward(configs)
    vehicle_params = {
        "vehicle_type": 2,  # 1: FORD_ESCORT; 2: BMW_320i; 3: VW_VANAGON
        "vehicle_model": 0  # 0: PM, 1: ST, 2: KS, 3: MB, 4: YawRate
    }
    ego_action: DiscretePMJerkAction = DiscretePMJerkAction(vehicle_params, 5, 5)
    observations = {
        "distance_goal_time": [-1],
    }

    assert math.isclose(reward.goal_time_reward(observations, ego_action), 0.5)

    observations = {
        "distance_goal_time": [1],
    }
    assert math.isclose(reward.goal_time_reward(observations, ego_action), 0)

    observations = {
        "distance_goal_time": [0],
    }
    assert math.isclose(reward.goal_time_reward(observations, ego_action), 0)


@unit_test
@functional
def test_goal_velocity_reward():
    reward = HybridReward(configs)
    observations = {
        "distance_goal_time": [0],
        "distance_goal_long": [0],
        "distance_goal_lat": [0],
        "distance_goal_velocity": [-1.]
    }

    class Dummy(object):
        pass

    ego_action = Dummy()
    setattr(ego_action, "vehicle", Dummy())
    setattr(ego_action.vehicle, "state", Dummy())
    setattr(ego_action.vehicle, "vehicle_model", VehicleModel.PM)
    setattr(ego_action.vehicle.state, "velocity", 2.)
    setattr(ego_action.vehicle.state, "velocity_y", 0.5)

    assert math.isclose(reward.goal_velocity_reward(observations, ego_action), 0.3606, abs_tol=1e-2)


@unit_test
@functional
def test_goal_orientation_reward():
    reward = HybridReward(configs)
    observations = {
        "distance_goal_time": [0],
        "distance_goal_long": [0],
        "distance_goal_lat": [0],
        "distance_goal_orientation": [1.]
    }
    assert math.isclose(reward.goal_orientation_reward(observations), (math.exp(-1.0 / np.pi) * 0.5))


@pytest.mark.parametrize(
    ("observations", "expected_reward"),
    [(
            {
                "v_ego": np.array([23., 0.]),
                "lane_based_v_rel": np.array([0., 0., 0., 0., -11.5, 0.]),
                "lane_based_p_rel": np.array([0., 0., 0., 0., 10., 0.]),
                "lane_change": np.array([0.])
            },
            -0.05510302167
    ),
        (
                {
                    "v_ego": np.array([23.]),
                    "lane_based_v_rel": np.array([0., 0., 0., 0., -11.5, 0.]),
                    "lane_based_p_rel": np.array([0., 0., 0., 0., 10., 0.]),
                    "lane_change": np.array([0.])
                },
                -0.05510302167
        ),
        (
                {
                    "v_ego": np.array([23.]),
                    "lane_based_v_rel": np.array([0., 0., 0., 0., 1., 0.]),
                    "lane_based_p_rel": np.array([0., 0., 0., 0., 10., 0.]),
                    "lane_change": np.array([0.])
                },
                0.
        ),
        (
                {
                    "v_ego": np.array([11.5]),
                    "lane_based_v_rel": np.array([0., -11.5, 0., 0., 0., 0.]),
                    "lane_based_p_rel": np.array([0., 10., 0., 0., 10., 0.]),
                    "lane_change": np.array([1.])
                },
                -0.05510302167
        )],
)
@unit_test
@functional
def test_safe_distance_reward(observations, expected_reward):
    configs["surrounding_configs"]["observe_lane_change"] = True
    configs["surrounding_configs"]["observe_lane_rect_surrounding"] = True
    configs["surrounding_configs"]["lane_rect_sensor_range_length"] = 100.
    configs["surrounding_configs"]["lane_rect_sensor_range_width"] = 8.
    configs["reward_configs"]["hybrid_reward"]["reward_safe_distance_coef"] = -1.

    reward = HybridReward(configs)

    assert math.isclose(reward.safe_distance_reward(observations), expected_reward)


@unit_test
@functional
def test_orientation_to_reference_reward():
    reward = HybridReward(configs)
    observations = {
        "route_reference_path_orientations": [0., 1., 0., 0., 0.]
    }
    assert math.isclose(reward.orientation_to_reference_reward(observations), (math.exp(-1.0 / np.pi) * 1))


@pytest.mark.parametrize(
    "benchmark_id",
    [  # "FRA_Miramas-5_1_T-1",
        "USA_US101-14_1_T-1"],
)
@module_test
@nonfunctional
def test_reward_calculation(benchmark_id):
    config_path = constants.PATH_PARAMS["configs"]["commonroad-v1"]
    resource_path = resource_root("test_reward")

    xml_scenarios_path = os.path.join(resource_path)
    output_base_path = os.path.join(output_root("test_reward"))

    pickle_path = os.path.join(output_base_path, "pickles")

    # Pickle CommonRoad scenarios
    pickle_xml_scenarios(
        input_dir=xml_scenarios_path,
        output_dir=pickle_path
    )

    meta_scenario_path = os.path.join(pickle_path, "meta_scenario")
    problem_path = os.path.join(pickle_path, "problem")

    with open(config_path, "r") as f:
        config = yaml.safe_load(f)
    config = config["env_configs"]
    config["vehicle_params"]["vehicle_model"] = 0
    reward = HybridReward(config)
    ego_action: DiscretePMJerkAction = DiscretePMJerkAction(config["vehicle_params"], 5, 5)
    observation_collector = ObservationCollector(config)

    with open(os.path.join(meta_scenario_path, "meta_scenario_reset_dict.pickle"), "rb") as f:
        meta_scenario_reset_dict = pickle.load(f)

    # benchmark_id = "FRA_Miramas-5_1_T-1"
    with open(os.path.join(problem_path, f"{benchmark_id}.pickle"), "rb") as f:
        problem_dict: dict = pickle.load(f)

    problem = list(problem_dict["planning_problem_set"].planning_problem_dict.values())[0]

    scenario_id = ScenarioID.from_benchmark_id(benchmark_id, "2020a")
    map_id = parse_map_name(scenario_id)
    reset_config = meta_scenario_reset_dict[map_id]
    scenario = restore_scenario(reset_config["meta_scenario"], problem_dict["obstacle"], scenario_id)
    ego_action.reset(problem.initial_state, scenario.dt)
    observation_collector.reset(scenario, problem, reset_config, benchmark_id)
    observation_collector.observe(ego_action.vehicle)
    reward.reset(observation_collector.observation_dict, ego_action)
    reward.calc_reward(observation_collector.observation_dict, ego_action)



"""
Module for drawing the ccosy
"""

__author__ = "Brian Liao, Niels Muendler, Peter Kocsis"
__copyright__ = "TUM Cyber-Physical System Group"
__credits__ = [""]
__version__ = "1.0"
__maintainer__ = "Xiao Wang"
__email__ = "xiao.wang@tum.de"
__status__ = "Released"

import argparse
import random
from typing import List

import commonroad_dc.pycrcc as pycrcc
import commonroad_dc.pycrccosy as pycrccosy
# import pycrcc
import imageio
# import commonroad_ccosy.visualization.draw_dispatch

import numpy as np

from commonroad.planning.planning_problem import PlanningProblem
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.scenario import Scenario
from commonroad.scenario.trajectory import State
from commonroad.visualization.draw_dispatch_cr import draw_object
from commonroad_dc.geometry.util import resample_polyline
from commonroad_dc.collision.visualization import draw_dispatch as crdc_draw_dispatch
from matplotlib import gridspec

from commonroad_rl.gym_commonroad.commonroad_env import CommonroadEnv, restore_scenario
from commonroad_rl.gym_commonroad.observation.goal_observation import GoalObservation
from commonroad_rl.gym_commonroad.action.vehicle import Vehicle

import matplotlib
# try:
#     matplotlib.use("TkAgg")
# except:
#     matplotlib.use("AGG")
import matplotlib.pyplot as plt

VEHICLE_PARAMS = {
    "vehicle_type": 2,  # VehicleType.BMW_320i
    "vehicle_model": 0,
}


def argsparser():
    global args
    parser = argparse.ArgumentParser()
    parser.add_argument("--draw", "-d", action="store_true")
    parser.add_argument("--animate", "-a", action="store_true")
    return parser.parse_args()


def get_lanelet_orientation_at_state(lanelet: Lanelet, state: State):
    """
    Approximates the lanelet orientation with the two closest point to the given state

    :param lanelet: Lanelet on which the orientation at the given state should be calculated
    :param state: State where the lanelet's orientation should be calculated
    :return: An orientation in interval [-pi,pi]
    """
    # TODO optimize more for speed

    center_vertices = lanelet.center_vertices

    position_diff = []
    for idx in range(len(center_vertices) - 1):
        vertex1 = center_vertices[idx]
        position_diff.append(np.linalg.norm(state.position - vertex1))

    closest_vertex_index = position_diff.index(min(position_diff))

    vertex1 = center_vertices[closest_vertex_index, :]
    vertex2 = center_vertices[closest_vertex_index + 1, :]
    direction_vector = vertex2 - vertex1
    return np.arctan2(direction_vector[1], direction_vector[0])


def draw_ccosy(
        scenario: Scenario,
        planning_problem: PlanningProblem,
        ccosy_list: List[pycrccosy.CurvilinearCoordinateSystem],
        ego_state: State = None,
):
    """
    Draws all ccosy to the current plot

    :param scenario: The scenario to be plotted
    :param planning_problem: The planning problem to be plotted
    :param ccosy_list: The list of the ccosies to be plotted
    :param ego_state: The state of the ego vehicle to be plotted (optional)
    """
    for global_cosy in ccosy_list:
        draw_object(
            global_cosy.get_segment_list()
        )
    draw_object(
        scenario,
        draw_params={
            "time_begin": 0,
            "scenario": {
                "lanelet_network": {
                    "lanelet": {
                        "show_label": False,
                        "fill_lanelet": False,
                    }
                },
            },
            "planning_problem_set": {
                "planning_problem": {
                    "goal_region": {
                        "draw_shape": True,
                        "shape": {
                            "polygon": {
                                "opacity": 0.5,
                                "linewidth": 0.5,
                                "facecolor": "#f1b514",
                                "edgecolor": "#302404",
                                "zorder": 15,
                            },
                            "rectangle": {
                                "opacity": 0.5,
                                "linewidth": 0.5,
                                "facecolor": "#f1b514",
                                "edgecolor": "#302404",
                                "zorder": 15,
                            },
                            "circle": {
                                "opacity": 0.5,
                                "linewidth": 0.5,
                                "facecolor": "#f1b514",
                                "edgecolor": "#302404",
                                "zorder": 15,
                            },
                        },
                    },
                },
            },
        },
    )
    draw_object(planning_problem)

    ego_vehicle = Vehicle.create_vehicle(VEHICLE_PARAMS)
    ego_params = ego_vehicle.params

    # Draw ego
    if ego_state is None:
        ego_state = planning_problem.initial_state
    collision_object = pycrcc.RectOBB(
        ego_params.l / 2,
        ego_params.w / 2,
        ego_state.orientation,
        ego_state.position[0],
        ego_state.position[1],
    )
    crdc_draw_dispatch.draw_object(
        collision_object, draw_params={"collision": {"facecolor": "red"}}
    )
    plt.gca().set_aspect("equal")


def animate_goal_obs(goal_obs: GoalObservation, title: str):
    """
    Creates animation about the ccosy and the ego vehicle moving over the ccosies
    Next to the scenario the goal related observations are plotted as well

    :param goal_obs: The GoalObservation object handles the goal related observations
    :param title: The title of the plot
    """
    # Create states over the route which will be animated
    states_to_test = []
    merged_lanelets = goal_obs.get_reference_lanelet_route()
    for lanelet in merged_lanelets:
        polyline = lanelet.center_vertices
        polyline = resample_polyline(polyline, step=2.0)

        # Add some point before and after the lane for better visualization
        normal_vector = polyline[1, :] - polyline[0, :]
        points_before_the_lane = [
            (polyline[0, :] - step * normal_vector) for step in reversed(range(1, 10))
        ]

        normal_vector = polyline[-1, :] - polyline[-2, :]
        points_after_the_lane = [
            (polyline[-1, :] + step * normal_vector) for step in range(1, 10)
        ]

        polyline = np.vstack((points_before_the_lane, polyline, points_after_the_lane))

        states_to_test.extend(
            [
                State(
                    position=position,
                    orientation=get_lanelet_orientation_at_state(
                        lanelet, State(position=position)
                    ),
                )
                for position in polyline
            ]
        )

    plots = []
    long_distances = []
    lat_distances = []
    long_lane_distances = []
    num_states = len(states_to_test)

    fig = plt.figure(figsize=(25, 10))
    fig.suptitle(title)

    # Plot each steps
    for state_to_test in states_to_test:
        long_dist, lat_dist = goal_obs.get_long_lat_distance_to_goal(
            state_to_test.position
        )
        long_lane_change_dist = goal_obs.get_long_distance_until_lane_change(
            state_to_test
        )

        long_distances.append(long_dist)
        lat_distances.append(lat_dist)
        long_lane_distances.append(long_lane_change_dist)

        fig.clf()
        # Left side of the animation: The scenario
        outer = gridspec.GridSpec(1, 2, wspace=0.2, hspace=0.2)

        inner = gridspec.GridSpecFromSubplotSpec(
            1, 1, subplot_spec=outer[0], wspace=0.1, hspace=0.1
        )
        ax = plt.Subplot(fig, inner[0])
        fig.add_subplot(ax)
        draw_ccosy(
            goal_obs.scenario,
            goal_obs.planning_problem,
            goal_obs.navigator.ccosy_list,
            state_to_test,
        )

        # Right side of the animation: The plots
        inner = gridspec.GridSpecFromSubplotSpec(
            3, 1, subplot_spec=outer[1], wspace=0.1, hspace=0.1
        )
        ax = plt.Subplot(fig, inner[0])
        ax.set_xlim([0, num_states - 1])
        ax.set_ylim([min([min(long_distances), 0]) - 0.1, max(long_distances) + 0.1])
        fig.add_subplot(ax)
        plt.plot(long_distances, label="d_long")
        ax.legend()

        ax = plt.Subplot(fig, inner[1])
        ax.set_xlim([0, num_states - 1])
        ax.set_ylim([min([min(lat_distances), 0]) - 0.1, max(lat_distances) + 0.1])
        fig.add_subplot(ax)
        plt.plot(lat_distances, label="d_lat")
        ax.legend()

        ax = plt.Subplot(fig, inner[2])
        ax.set_xlim([0, num_states - 1])
        ax.set_ylim(
            [min([min(long_lane_distances), 0]) - 0.1, max(long_lane_distances) + 0.1]
        )
        fig.add_subplot(ax)
        plt.plot(long_lane_distances, label="d_long_lane")
        ax.legend()

        fig.canvas.draw()  # draw the canvas, cache the renderer
        image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
        image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        plots.append(image)

    # Create gif
    imageio.mimsave(f"./goal_dist_{title}.gif", plots, fps=5)


def main(arguments):
    import commonroad_rl
    print(commonroad_rl.__file__)
    cr_env = CommonroadEnv()

    all_problems = list(cr_env.all_problem_dict.items())
    num_all_problems = len(all_problems)
    for i, (benchmark_id, problem_dict) in enumerate(all_problems):

        meta_scenario = cr_env.problem_meta_scenario_dict[benchmark_id]
        obstacle_list = problem_dict["obstacle"]
        scenario = restore_scenario(meta_scenario, obstacle_list)

        planning_problems = list(
            problem_dict["planning_problem_set"].planning_problem_dict.values()
        )
        num_planning_problems = len(planning_problems)
        if num_planning_problems > 1:
            print(
                f"There are {num_planning_problems} planning problems in the scenario {benchmark_id}"
            )

        planning_problem = random.choice(planning_problems)

        try:
            goal_obs = GoalObservation(scenario, planning_problem)

            state_to_test = planning_problem.initial_state
            long_lat_dist = goal_obs.get_long_lat_distance_to_goal(
                state_to_test.position
            )
            long_lane_change_dist = goal_obs.get_long_distance_until_lane_change(
                state_to_test
            )

            print(
                f"{i + 1}/{num_all_problems} DONE {benchmark_id} - "
                f"Long-lat distance: {long_lat_dist}, Lane change dist: {long_lane_change_dist}"
            )
        except ValueError as ex:
            print(
                f"{i + 1}/{num_all_problems} ------------- ERROR ------------ in {benchmark_id}:{ex}"
            )
            raise ex

        if arguments.animate:
            animate_goal_obs(goal_obs, benchmark_id)

        if arguments.draw:
            fig = plt.figure(figsize=(25, 10))
            fig.suptitle(benchmark_id)
            draw_ccosy(scenario, planning_problem, goal_obs.navigator.ccosy_list)
            plt.show()


if __name__ == "__main__":
    args = argsparser()
    main(args)



from collections import defaultdict, OrderedDict
from typing import Union, Dict, List, Tuple
import gym
import numpy as np
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.obstacle import State
from commonroad.scenario.scenario import Scenario
from commonroad.visualization.mp_renderer import MPRenderer, ZOrders
from commonroad.visualization.util import LineDataUnits
from commonroad_rl.gym_commonroad.action.vehicle import Vehicle
from commonroad_rl.gym_commonroad.observation.observation import Observation
from commonroad_rl.gym_commonroad.observation.goal_observation import GoalObservation
from commonroad_rl.gym_commonroad.utils.navigator import Navigator
from numpy import ndarray
import commonroad_dc.pycrcc as pycrcc
from commonroad_dc.geometry.geometry import CurvilinearCoordinateSystem
from shapely.geometry import Point, LineString

from commonroad_rl.gym_commonroad.utils.scenario import approx_orientation_vector, get_lane_marker


class LaneletNetworkObservation(Observation):
    def __init__(self, configs: Dict, configs_name: str = "lanelet_configs"):
        # Read configs
        configs = configs[configs_name]
        self.strict_off_road_check: bool = configs.get("strict_off_road_check")
        self.non_strict_check_circle_radius: float = configs.get("non_strict_check_circle_radius")

        self.observe_lat_offset: bool = configs.get("observe_lat_offset")
        self.observe_left_marker_distance: bool = configs.get("observe_left_marker_distance")
        self.observe_right_marker_distance: bool = configs.get("observe_right_marker_distance")
        self.observe_left_road_edge_distance: bool = configs.get("observe_left_road_edge_distance")
        self.observe_right_road_edge_distance: bool = configs.get("observe_right_road_edge_distance")
        self.observe_is_off_road: bool = configs.get("observe_is_off_road")
        self.observe_lane_curvature: bool = configs.get("observe_lane_curvature")

        # extrapolated_positions
        self.observe_static_extrapolated_positions: bool = configs.get("observe_static_extrapolated_positions")
        self.static_extrapolation_samples: List[float] = configs.get("static_extrapolation_samples")
        self.observe_dynamic_extrapolated_positions: bool = configs.get("observe_dynamic_extrapolated_positions")
        self.dynamic_extrapolation_samples: List[float] = configs.get("dynamic_extrapolation_samples")

        # reference path waypoints configs
        self.observe_route_reference_path: bool = configs.get("observe_route_reference_path")
        self.distances_route_reference_path: List[float] = configs.get("distances_route_reference_path")
        # reference lanelets waypoints configs
        self.observe_route_multilanelet_waypoints: bool = configs.get("observe_route_multilanelet_waypoints")
        self.distances_and_ids_multilanelet_waypoints: Tuple[List[float], List[int]] = configs.get("distances_and_ids_multilanelet_waypoints")
        # distance to reference path
        self.observe_distance_togoal_via_referencepath: bool = configs.get("observe_distance_togoal_via_referencepath")

        self.observation_dict = OrderedDict()
        self.observation_history_dict = defaultdict(list)

    def build_observation_space(self) -> OrderedDict:
        """ builds observation space for LaneletNetworkObservation """
        observation_space_dict = OrderedDict()

        if self.observe_is_off_road:
            observation_space_dict["is_off_road"] = gym.spaces.Box(0, 1, (1,), dtype=np.int8)
        if self.observe_left_marker_distance:
            observation_space_dict["left_marker_distance"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_right_marker_distance:
            observation_space_dict["right_marker_distance"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_left_road_edge_distance:
            observation_space_dict["left_road_edge_distance"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_right_road_edge_distance:
            observation_space_dict["right_road_edge_distance"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_lat_offset:
            observation_space_dict["lat_offset"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)
        if self.observe_lane_curvature:
            observation_space_dict["lane_curvature"] = gym.spaces.Box(-np.inf, np.inf, (1,), dtype=np.float32)

        if self.observe_dynamic_extrapolated_positions:
            sampling_points = self.dynamic_extrapolation_samples
            observation_space_dict["extrapolation_dynamic_off"] = gym.spaces.Box(-np.inf, np.inf,
                                                                                 (len(sampling_points),),
                                                                                 dtype=np.float32)
        if self.observe_static_extrapolated_positions:
            sampling_points = self.static_extrapolation_samples
            observation_space_dict["extrapolation_static_off"] = gym.spaces.Box(-np.inf, np.inf,
                                                                                (len(sampling_points),),
                                                                                dtype=np.float32)
        if self.observe_route_reference_path:
            distances = self.distances_route_reference_path
            assert sorted(set(distances)) == sorted(distances) and all(isinstance(x, int) or isinstance(x, float) for x in distances), \
                "the config setting distances_and_ids_multilanelet_waypoints[0] / distances must be a set of floats"
            observation_space_dict["route_reference_path_positions"] = gym.spaces.Box(
                -np.inf, np.inf, (len(distances) * 2,), dtype=np.float32
            )
            observation_space_dict["route_reference_path_orientations"] = gym.spaces.Box(
                -np.inf, np.inf, (len(distances),), dtype=np.float32
            )
        if self.observe_route_multilanelet_waypoints:
            distances, ids = self.distances_and_ids_multilanelet_waypoints
            assert sorted(set(distances)) == sorted(distances) and all(isinstance(x, int) or isinstance(x, float) for x in distances), \
                "the config setting distances_and_ids_multilanelet_waypoints[0] / distances must be a set of float/int"
            assert sorted(set(ids)) == sorted(ids) and all(isinstance(x, int) for x in ids), \
                "the config setting distances_and_ids_multilanelet_waypoints[1] / relative ids must be a set of ints"

            observation_space_dict["route_multilanelet_waypoints_positions"] = gym.spaces.Box(-np.inf, np.inf,
                                                                                (len(ids) * len(distances) * 2,),
                                                                                dtype=np.float32)
            observation_space_dict["route_multilanelet_waypoints_orientations"] = gym.spaces.Box(-np.inf, np.inf,
                                                                                (len(ids) * len(distances), ),
                                                                                dtype=np.float32)
        if self.observe_distance_togoal_via_referencepath:
            observation_space_dict["distance_togoal_via_referencepath"] = gym.spaces.Box(-np.inf, np.inf,
                                                                                (3,),
                                                                                dtype=np.float32)

        return observation_space_dict

    def observe(self, scenario: Scenario, ego_vehicle: Vehicle, ego_lanelet: Lanelet, road_edge: dict,
                local_ccosy: Union[None, CurvilinearCoordinateSystem] = None, navigator: Union[Navigator, None] = None
                ) -> Union[ndarray, Dict]:
        self._scenario = scenario

        # Check if the ego vehicle is off road
        is_off_road = self._check_is_off_road(ego_vehicle, road_edge)

        if self.observe_is_off_road:
            self.observation_history_dict["is_off_road"].append(is_off_road)
            self.observation_dict["is_off_road"] = np.array([is_off_road])

        if any((self.observe_left_marker_distance, self.observe_right_marker_distance,
                self.observe_left_road_edge_distance, self.observe_right_road_edge_distance)):
            self._get_distance_to_marker_and_road_edge(ego_vehicle.state, ego_lanelet, road_edge)

        # TODO: merge two functions to reduce run time
        if self.observe_lat_offset:
            self._get_lat_offset(ego_vehicle.state, local_ccosy)

        if self.observe_lane_curvature:
            self._get_lane_curvature(ego_vehicle.state.position, local_ccosy)

        # Get the relative offset of current and future positions from center vertices TODO
        if self.observe_static_extrapolated_positions:
            sampling_points = self.static_extrapolation_samples
            static_lat_offset, static_pos = self._get_relative_future_goal_offsets(ego_vehicle.state, sampling_points,
                                                                                   static=True, navigator=navigator)

            self._extrapolation_static_off = (np.array(static_lat_offset))
            self._extrapolation_static_pos = (np.array(static_pos))
            self.observation_dict["extrapolation_static_off"] = np.array(static_lat_offset)

        if self.observe_dynamic_extrapolated_positions:
            sampling_points = self.dynamic_extrapolation_samples
            dynamic_lat_offset, dynamic_pos = self._get_relative_future_goal_offsets(ego_vehicle.state, sampling_points,
                                                                                     static=False, navigator=navigator)
            self._extrapolation_dynamic_off = (np.array(dynamic_lat_offset))
            self._extrapolation_dynamic_pos = (np.array(dynamic_pos))
            self.observation_dict["extrapolation_dynamic_off"] = np.array(dynamic_lat_offset)

        # get waypoints of the reference path
        if self.observe_route_reference_path:
            pos, orient = navigator.get_waypoints_of_reference_path(
                ego_vehicle.state, distances_ref_path=self.distances_route_reference_path,
                observation_cos=Navigator.CosyVehicleObservation.AUTOMATIC
            )
            assert pos.shape == (len(self.distances_route_reference_path), 2)
            assert orient.shape == (len(self.distances_route_reference_path),)
            # assert orient.shape == (len(self.distances_route_reference_path),1), "assert works"
            self.observation_dict["route_reference_path_positions"] = pos.flatten()
            self.observation_dict["route_reference_path_orientations"] = orient.flatten()

        if self.observe_route_multilanelet_waypoints:
            distances, ids = self.distances_and_ids_multilanelet_waypoints

            pos, orient = navigator.get_referencepath_multilanelets_waypoints(
                ego_vehicle.state, distances_per_lanelet=distances, lanelets_id_rel=ids,
                observation_cos=Navigator.CosyVehicleObservation.AUTOMATIC
            )
            self.observation_dict["route_multilanelet_waypoints_positions"] = pos.flatten()
            self.observation_dict["route_multilanelet_waypoints_orientations"] = orient.flatten()
        
        if self.observe_distance_togoal_via_referencepath:
            distance_long, distance_lat, indomain = navigator.get_longlat_togoal_on_reference_path(ego_vehicle.state)
            self.observation_dict["distance_togoal_via_referencepath"] = np.array((distance_long, distance_lat, indomain))

        return self.observation_dict

    def draw(self, render_configs: Dict, render: MPRenderer, ego_vehicle: Vehicle, road_edge: Union[None, Dict] = None,
             ego_lanelet: Union[Lanelet, None] = None, navigator: Union[Navigator, None] = None):
        """ Method to draw the observation """
        # Draw road boundaries
        if render_configs["render_road_boundaries"]:
            road_edge["boundary_collision_object"].draw(render)
        # Plot ego lanelet center vertices
        if render_configs["render_ego_lanelet_center_vertices"]:
            line = LineDataUnits(ego_lanelet.center_vertices[:, 0], ego_lanelet.center_vertices[:, 1],
                                 zorder=ZOrders.LANELET_LABEL, markersize=1., color="pink",
                                 marker="x", label="Ego center vertices")
            render.dynamic_artists.append(line)

        # Extrapolated future positions
        if self.observe_static_extrapolated_positions and render_configs["render_static_extrapolated_positions"]:
            for future_pos in self._extrapolation_static_pos:
                render.dynamic_artists.append(
                    LineDataUnits(future_pos[0], future_pos[1], color="r", marker="x",
                                  zorder=21, label="static_extrapolated_positions"))

        if self.observe_dynamic_extrapolated_positions and render_configs["render_dynamic_extrapolated_positions"]:
            for future_pos in self._extrapolation_dynamic_pos:
                render.dynamic_artists.append(
                    LineDataUnits(future_pos[0], future_pos[1], color="b", marker="x",
                                  zorder=21, label="dynamic_extrapolated_positions"))

        # Plot Navigator Observations in non-local (global) CoSy
        if self.observe_route_reference_path and render_configs["render_ccosy_nav_observations"]:
            pos, _ = navigator.get_waypoints_of_reference_path(
                ego_vehicle.state, distances_ref_path=self.distances_route_reference_path,
                observation_cos=Navigator.CosyVehicleObservation.LOCALCARTESIAN
                # coordinate axes for plot in direction of global CoSy
            )
            pos_global = ego_vehicle.state.position + pos  # back to non local CoSy Origin
            render.dynamic_artists.append(
                LineDataUnits(pos_global[:, 0], pos_global[:, 1], zorder=22, marker='v',
                              color='yellow', label="ccosy_nav_observations"))

        if self.observe_route_multilanelet_waypoints and render_configs["render_ccosy_nav_observations"]:
            distances, ids = self.distances_and_ids_multilanelet_waypoints

            pos, _ = navigator.get_referencepath_multilanelets_waypoints(
                ego_vehicle.state, distances_per_lanelet=distances, lanelets_id_rel=ids,
                observation_cos=Navigator.CosyVehicleObservation.LOCALCARTESIAN
                # coordinate axes for plot in direction of global CoSy
            )
            for po in pos:
                po_global = ego_vehicle.state.position + po  # back to non local CoSy Origin
                render.dynamic_artists.append(
                    LineDataUnits(po_global[:, 0], po_global[:, 1], zorder=22, marker='v',
                                  color='purple', label="ccosy_nav_observations"))

    def _get_relative_future_goal_offsets(self, ego_state: State, sampling_points: List[float], static: bool,
                                          navigator: Navigator) -> Tuple[List[float], List[ndarray]]:
        """
        Get the relative offset of current and future positions from center vertices. Positive if left.
        For a given static extrapolation, the future position at "static" m/s after sampling_points seconds is given.
        For static = True this means the future position in exactly sampling_points meters.
        Otherwise for static = False, the future position at the current velocity after sampling_points seconds is
        given.

        :param ego_state: State of ego vehicle
        :param sampling_points: Parameter of evaluating the future position, see description above
        :param static: Curvilinear coordinate system
        :param navigator: Navigator of current planning problem
        :return: Offset of step_parameter future positions as well as the positions themselves
        """

        ego_state_orientation = ego_state.orientation if hasattr(ego_state, "orientation") else np.arctan2(
            ego_state.velocity_y, ego_state.velocity)
        v = approx_orientation_vector(ego_state_orientation) * (ego_state.velocity if static is False else 1.0)

        # quadratic steps may make sense based on the fact that braking distance is
        # proportional to the velocity squared (https://en.wikipedia.org/wiki/Braking_distance)
        positions = [ego_state.position + (v * i) for i in sampling_points]
        lat_offset = [GoalObservation.get_long_lat_distance_to_goal(p, navigator)[1] for p in positions]

        # Fix nans for positions outside the ccosy
        for i in range(len(lat_offset)):
            if np.isnan(lat_offset[i]):
                extrapolation_off = self._extrapolation_static_off if static else self._extrapolation_dynamic_off
                if len(extrapolation_off) > 0:
                    # We can assume that the ego vehicle starts inside the ccosy,
                    # however not that any extrapolated position is inside the ccosy
                    lat_offset[i] = extrapolation_off[i]
                else:
                    lat_offset[i] = 0

        return lat_offset, positions

    def _check_is_off_road(self, ego_vehicle: Vehicle, road_edge: dict) -> bool:
        """
        Check if the ego vehicle is off road.
        """
        strict_off_road_check = self.strict_off_road_check
        non_strict_check_circle_radius = self.non_strict_check_circle_radius if not self.strict_off_road_check else None

        collision_ego_vehicle = ego_vehicle.collision_object if strict_off_road_check else pycrcc.Circle(
            non_strict_check_circle_radius, ego_vehicle.state.position[0], ego_vehicle.state.position[1])

        is_off_road = collision_ego_vehicle.collide(road_edge["boundary_collision_object"])

        return is_off_road

    def _get_distance_to_marker_and_road_edge(self, ego_state: State, ego_lanelet: Lanelet, road_edge: dict):

        left_marker_line, right_marker_line = get_lane_marker(ego_lanelet)

        current_left_road_edge, current_right_road_edge = self._get_road_edge(road_edge, ego_lanelet.lanelet_id)

        (left_marker_distance, right_marker_distance, left_road_edge_distance, right_road_edge_distance) \
            = LaneletNetworkObservation.get_distance_to_marker_and_road_edge(ego_state, left_marker_line,
                                                                             right_marker_line, current_left_road_edge,
                                                                             current_right_road_edge)

        self.observation_dict["left_marker_distance"] = np.array([left_marker_distance])
        self.observation_dict["right_marker_distance"] = np.array([right_marker_distance])
        self.observation_dict["left_road_edge_distance"] = np.array([left_road_edge_distance])
        self.observation_dict["right_road_edge_distance"] = np.array([right_road_edge_distance])

    def _get_lane_curvature(self, ego_position: np.array, local_ccosy: CurvilinearCoordinateSystem):
        lane_curvature = LaneletNetworkObservation.get_lane_curvature(ego_position, local_ccosy)
        if np.isnan(lane_curvature):
            assert len(self.observation_history_dict["lane_curvature"]) > 0, \
                "Ego vehicle started outside the local coordinate system"
            lane_curvature = self.observation_history_dict["lane_curvature"][-1]
        self.observation_history_dict["lane_curvature"].append(lane_curvature)
        self.observation_dict["lane_curvature"] = np.array([lane_curvature])

    def _get_lat_offset(self, ego_state: State, local_ccosy: CurvilinearCoordinateSystem):

        lat_offset = LaneletNetworkObservation.get_relative_offset(local_ccosy, ego_state.position)
        if np.isnan(lat_offset):
            # we assume that the ego vehicle starts inside the curvi_cosy
            assert (len(self.observation_history_dict["lat_offset"]) > 0), \
                "Ego vehicle started outside the local coordinate system"
            lat_offset = self.observation_history_dict["lat_offset"][-1]

        self.observation_history_dict["lat_offset"].append(lat_offset)
        self.observation_dict["lat_offset"] = np.array([lat_offset])

    def _get_road_edge(self, road_edge: dict, ego_vehicle_lanelet_id: int) -> Tuple[LineString, LineString]:
        """
        Get the left and right road edge of ego vehicle lanelet.

        :param ego_vehicle_lanelet_id: id of ego vehicle lanelet
        :return: left and right road edge
        """
        left_most_lanelet_id = road_edge["left_road_edge_lanelet_id_dict"][ego_vehicle_lanelet_id]
        right_most_lanelet_id = road_edge["right_road_edge_lanelet_id_dict"][ego_vehicle_lanelet_id]
        left_road_edge = road_edge["left_road_edge_dict"][left_most_lanelet_id]
        right_road_edge = road_edge["right_road_edge_dict"][right_most_lanelet_id]
        return left_road_edge, right_road_edge

    @staticmethod
    def get_relative_offset(curvi_cosy: CurvilinearCoordinateSystem, position: np.ndarray) -> float:
        """
        Get the relative offset of ego vehicle from center vertices. Positive if left.

        :param curvi_cosy: curvilinear coordinate system
        :param position: The position of the ego vehicle
        :return: offset
        """
        try:
            _, ego_vehicle_lat_position = curvi_cosy.convert_to_curvilinear_coords(position[0], position[1])
        except ValueError:
            ego_vehicle_lat_position = np.nan

        return ego_vehicle_lat_position

    @staticmethod
    def get_distance_to_marker_and_road_edge(ego_vehicle_state: State, left_marker_line: LineString,
                                             right_marker_line: LineString, left_road_edge: LineString,
                                             right_road_edge: LineString, ) -> Tuple[float, float, float, float]:
        """
        Get the distance to lane markers and the road edge

        :param ego_vehicle_state: The state of the ego vehicle
        :param left_marker_line: The left marker line
        :param right_marker_line: The right marker line
        :param left_road_edge: The left road edge
        :param right_road_edge: The right road edge
        :return: Tuple of the distances to the left marker, right marker, left road edge and right road edge
        """
        ego_vehicle_point = Point(ego_vehicle_state.position[0], ego_vehicle_state.position[1])

        distance_left_marker = LaneletNetworkObservation.get_distance_point_to_linestring(ego_vehicle_point,
                                                                                          left_marker_line)
        distance_right_marker = LaneletNetworkObservation.get_distance_point_to_linestring(ego_vehicle_point,
                                                                                           right_marker_line)
        distance_left_road_edge = LaneletNetworkObservation.get_distance_point_to_linestring(ego_vehicle_point,
                                                                                             left_road_edge)
        distance_right_road_edge = LaneletNetworkObservation.get_distance_point_to_linestring(ego_vehicle_point,
                                                                                              right_road_edge)

        return distance_left_marker, distance_right_marker, distance_left_road_edge, distance_right_road_edge

    @staticmethod
    def get_distance_point_to_linestring(p: Point, line: LineString) -> float:
        """
        Get the distance of a point to the given line

        :param p: The point
        :param line: The line
        :return: The distance between the point and the line
        """
        nearest_point = line.interpolate(line.project(p))
        return nearest_point.distance(p)

    @staticmethod
    def get_lane_curvature(ego_position: np.array, ccosy: CurvilinearCoordinateSystem):
        try:
            s, _ = ccosy.convert_to_curvilinear_coords(ego_position[0], ego_position[1])
        except ValueError:
            return np.nan

        return np.interp(s, ccosy.ref_pos, ccosy.ref_curv)


if __name__ == "__main__":
    import yaml
    from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

    config_file = PATH_PARAMS["configs"]["commonroad-v1"]
    with open(config_file, "r") as config_file:
        config = yaml.safe_load(config_file)
    configs = config["env_configs"]
    lanelet_network_observation = LaneletNetworkObservation(configs)
    print(lanelet_network_observation)



"""Unit tests for sparse reward class"""
from commonroad_rl.gym_commonroad.reward.sparse_reward import SparseReward
from commonroad_rl.tests.common.marker import unit_test, functional

configs = {
    "reward_type": "sparse_reward",
    "reward_configs":{"sparse_reward": {
        "reward_goal_reached": 50.,
        "reward_collision": -50.,
        "reward_off_road": -20.,
        "reward_friction_violation": -30.,
        "reward_time_out": -10.
    }}
}


@unit_test
@functional
def test_sparse_reward():
    reward = SparseReward(configs)
    observations = {
        "is_goal_reached": [True],
        "is_collision": [False],
        "is_off_road": [False],
        "is_friction_violation": [False],
        "is_time_out": [False]
    }

    assert reward.calc_reward(observations, None) == 50

    observations = {
        "is_goal_reached": [False],
        "is_collision": [False],
        "is_off_road": [True],
        "is_friction_violation": [False],
        "is_time_out": [False]
    }

    assert reward.calc_reward(observations, None) == -20



# TODO: delete or merge to duplicate files
"""
Module for dividing dataset into train and test set
"""

import argparse
import glob
import ntpath
import os
import random
from shutil import copyfile


def get_args():
    parser = argparse.ArgumentParser(
        description="Divide pickle files into training and testing"
    )
    parser.add_argument(
        "--input_dir", "-i", type=str, default="/home/chenx/project/pickles/problem"
    )
    parser.add_argument(
        "--output_dir_train",
        "-otrain",
        type=str,
        default="/home/chenx/project/pickles/problem_train",
    )
    parser.add_argument(
        "--output_dir_test",
        "-otest",
        type=str,
        default="/home/chenx/project/pickles/problem_test",
    )
    parser.add_argument("--seed", "-s", type=int, default=5)
    parser.add_argument("--train_ratio", "-tr_r", type=float, default=0.7)

    return parser.parse_args()


def main():
    args = get_args()
    fns = glob.glob(os.path.join(args.input_dir, "*.pickle"))
    random.seed(args.seed)
    random.shuffle(fns)

    train_path = args.output_dir_train
    test_path = args.output_dir_test

    os.makedirs(train_path, exist_ok=True)
    os.makedirs(test_path, exist_ok=True)

    fns = sorted(glob.glob(os.path.join(args.input_dir, "*.pickle")))
    random.shuffle(fns)

    if args.train_ratio == -1:
        num_train=len(fns)
        print("Copying training and testing data ...")
        for i, fn in enumerate(fns):
            print(f"{i + 1}/{num_train}", end="\r")
            copyfile(fn, os.path.join(train_path, ntpath.basename(fn)))
            copyfile(fn, os.path.join(test_path, ntpath.basename(fn)))
    else:
        num_train = int(len(fns) * args.train_ratio)

        print("Copying training data ...")
        for i, fn in enumerate(fns[:num_train]):
            print(f"{i + 1}/{num_train}", end="\r")
            copyfile(fn, os.path.join(train_path, ntpath.basename(fn)))

        print("Copying test data...")
        for i, fn in enumerate(fns[num_train:]):
            print(f"{i + 1}/{len(fns) - num_train}", end="\r")
            copyfile(fn, os.path.join(test_path, ntpath.basename(fn)))


if __name__ == "__main__":
    main()



import numpy as np

from stable_baselines.common.noise import ActionNoise


class LinearNormalActionNoise(ActionNoise):
    """
    A gaussian action noise with linear decay for the standard deviation.

    :param mean: (np.ndarray) the mean value of the noise
    :param sigma: (np.ndarray) the scale of the noise (std here)
    :param max_steps: (int)
    :param final_sigma: (np.ndarray)
    """

    def __init__(self, mean, sigma, max_steps, final_sigma=None):
        self._mu = mean
        self._sigma = sigma
        self._step = 0
        self._max_steps = max_steps
        if final_sigma is None:
            final_sigma = np.zeros_like(sigma)
        self._final_sigma = final_sigma

    def __call__(self):
        t = min(1.0, self._step / self._max_steps)
        sigma = (1 - t) * self._sigma + t * self._final_sigma
        self._step += 1
        return np.random.normal(self._mu, sigma)



"""
Module for generating CommonRoad scenarios using random actions
"""
import argparse
import os
import pickle
import multiprocessing

import gym
from commonroad.common.file_writer import CommonRoadFileWriter, OverwriteExistingFile
from commonroad.geometry.shape import Rectangle
from commonroad.planning.planning_problem import PlanningProblemSet
from commonroad.prediction.prediction import TrajectoryPrediction
from commonroad.scenario.obstacle import ObstacleType, DynamicObstacle
from commonroad.scenario.scenario import Tag
from commonroad.scenario.trajectory import Trajectory

from commonroad_rl.gym_commonroad.constants import PATH_PARAMS

os.environ["KMP_WARNINGS"] = "off"


def get_parser():
    parser = argparse.ArgumentParser(description="Generates scenarios using random actions",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--meta_path", type=str, default=PATH_PARAMS["meta_scenario"],
                        help="Path to pickled meta scenarios")
    parser.add_argument("--scenario_path", type=str, default=PATH_PARAMS["test_reset_config"],
                        help="Path to pickled test scenarios")
    parser.add_argument("--output_path", "-o", type=str, default="random_scenarios",
                        help="Path to pickled test scenarios")
    parser.add_argument("--n_cpu", "-n", type=int, default=1, help="Number of cpu processes")
    parser.add_argument("--create_collision_action", "-collision", action="store_true",
        help="Store actions that caused collision")

    return parser


def create_scenarios(args):
    env = gym.make("commonroad-v1",
                   meta_scenario_path=args.meta_path,
                   test_reset_config_path=args.scenario_path,
                   logging_path=None,
                   play=True)
    env.reset()
    initial_state = env.ego_action.vehicle.initial_state
    while True:
        action = env.action_space.sample()
        obs, reward, done, info = env.step(action)
        if done:
            # Append ego trajectory
            ego_trajectory = Trajectory(1, env.ego_action.vehicle.state_list)
            ego_shape = Rectangle(width=env.ego_action.vehicle.parameters.w, length=env.ego_action.vehicle.parameters.l)
            ego_id = env.scenario.generate_object_id()
            ego_type = ObstacleType.CAR
            ego_prediction = TrajectoryPrediction(ego_trajectory, ego_shape)
            ego_dynamic_obstacle = DynamicObstacle(ego_id,
                                                   ego_type,
                                                   ego_shape,
                                                   initial_state,
                                                   ego_prediction)
            # Add ego to scenario
            env.scenario.add_objects(ego_dynamic_obstacle)
            # Write ego trajectory to scenario
            author = 'Xiao Wang'
            affiliation = 'Technical University of Munich, Germany'
            source = 'highD + random action for ego'
            tags = {Tag.INTERSTATE, Tag.MULTI_LANE, Tag.NO_ONCOMING_TRAFFIC, Tag.HIGHWAY}

            planning_problem_set = PlanningProblemSet([env.planning_problem])
            fw = CommonRoadFileWriter(env.scenario, planning_problem_set, author, affiliation, source, tags)
            filename = os.path.join(args.output_path, f"{env.benchmark_id}-{ego_id}.xml")
            print(filename)
            fw.write_to_file(filename, OverwriteExistingFile.ALWAYS)

            # Reset env
            env.reset()


def create_collisions_multiple_process(args):
    if args.n_cpu <= 1:
        results = create_collisions_single_process(args.meta_path, args.scenario_path, seed=1)
    else:
        with multiprocessing.Pool(processes=args.n_cpu) as pool:
            results = pool.starmap(
                create_collisions_single_process,
                [
                    (args.meta_path, os.path.join(args.scenario_path, str(i)), i+1) for i in range(args.n_cpu)
                ]
            )
    with open(os.path.join(args.output_path, "collision_actions.pkl"), "wb") as f:
        pickle.dump(results, f)


def create_collisions_single_process(meta_path, scenario_path, seed):
    env = gym.make("commonroad-v1",
                   meta_scenario_path=meta_path,
                   test_reset_config_path=scenario_path,
                   logging_path=None,
                   play=True)
    env.seed(seed)
    env.reset()
    initial_state = env.ego_action.vehicle.initial_state
    log_actions = {}
    actions = []
    while True:
        action = env.action_space.sample()
        actions.append(action)
        # actions[str(env.scenario.scenario_id)].append(action)
        obs, reward, done, info = env.step(action)
        if done:
            if info["is_collision"]:
                log_actions[str(env.scenario.scenario_id)] = actions

            # Reset env
            try:
                env.reset()
                actions = []
            except IndexError:
                return log_actions



if __name__ == "__main__":
    args = get_parser().parse_args()
    os.makedirs(args.output_path, exist_ok=True)
    if args.create_collision_action:
        create_collisions_multiple_process(args)
    else:
        create_scenarios(args)



import os
import gym
import numpy as np
from commonroad_rl.tests.common.marker import *
from commonroad_rl.utils_run.wrappers import IncreaseTimeStepWrapper
from commonroad_rl.tests.common.path import resource_root, output_root
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios

resource_path = resource_root("test_wrapper")
pickle_xml_scenarios(
    input_dir=os.path.join(resource_path),
    output_dir=os.path.join(resource_path, "pickles")
)

meta_scenario_path = os.path.join(resource_path, "pickles", "meta_scenario")
problem_path = os.path.join(resource_path, "pickles", "problem")

@pytest.mark.parametrize(
    "planning_horizon",
    [1.,
     2.,
     0.4]
)

@module_test
@functional
def test_step(planning_horizon):
    env = gym.make("commonroad-v1",
                   meta_scenario_path=meta_scenario_path,
                   train_reset_config_path=problem_path,
                   test_reset_config_path=problem_path,
                   action_configs={"planning_horizon": planning_horizon})

    wrapper = IncreaseTimeStepWrapper
    env.seed(1)
    env = wrapper(env)
    env.reset()
    done = False
    step = 0
    while not done:
        action = np.array([0., 0.])
        # action = env.action_space.sample()
        obs, reward, done, info = env.step(action)
        step += 1
    expected_step = env.current_step // int(planning_horizon / env.scenario.dt) + 1
    assert step == expected_step



import gym
import numpy as np
from gym.wrappers import TimeLimit


class DoneOnSuccessWrapper(gym.Wrapper):
    """
    Reset on success and offsets the reward.
    Useful for GoalEnv.
    """

    def __init__(self, env, reward_offset=1.0):
        super(DoneOnSuccessWrapper, self).__init__(env)
        self.reward_offset = reward_offset

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        done = done or info.get("is_success", False)
        reward += self.reward_offset
        return obs, reward, done, info

    def compute_reward(self, achieved_goal, desired_goal, info):
        reward = self.env.compute_reward(achieved_goal, desired_goal, info)
        return reward + self.reward_offset


class TimeFeatureWrapper(gym.Wrapper):
    """
    Add remaining time to observation space for fixed length episodes.
    See https://arxiv.org/abs/1712.00378 and https://github.com/aravindr93/mjrl/issues/13.

    :param env: (gym.Env)
    :param max_steps: (int) Max number of steps of an episode
        if it is not wrapped in a TimeLimit object.
    :param test_mode: (bool) In test mode, the time feature is constant,
        equal to zero. This allow to check that the agent did not overfit this feature,
        learning a deterministic pre-defined sequence of actions.
    """

    def __init__(self, env, max_steps=1000, test_mode=False):
        assert isinstance(env.observation_space, gym.spaces.Box)
        # Add a time feature to the observation
        low, high = env.observation_space.low, env.observation_space.high
        low, high = np.concatenate((low, [0])), np.concatenate((high, [1.0]))
        env.observation_space = gym.spaces.Box(low=low, high=high, dtype=np.float32)

        super(TimeFeatureWrapper, self).__init__(env)

        if isinstance(env, TimeLimit):
            self._max_steps = env._max_episode_steps
        else:
            self._max_steps = max_steps
        self._current_step = 0
        self._test_mode = test_mode

    def reset(self):
        self._current_step = 0
        return self._get_obs(self.env.reset())

    def step(self, action):
        self._current_step += 1
        obs, reward, done, info = self.env.step(action)
        return self._get_obs(obs), reward, done, info

    def _get_obs(self, obs):
        """
        Concatenate the time feature to the current observation.

        :param obs: (np.ndarray)
        :return: (np.ndarray)
        """
        # Remaining time is more general
        time_feature = 1 - (self._current_step / self._max_steps)
        if self._test_mode:
            time_feature = 1.0
        # Optionnaly: concatenate [time_feature, time_feature ** 2]
        return np.concatenate((obs, [time_feature]))


class IncreaseTimeStepWrapper(gym.Wrapper):
    """
    Wrapper for increasing the time step size of the environment
    """
    def __init__(self, env):
        """
        initialization for wrapper
        :param env: environment to wrap
        """
        super().__init__(env)
        self.env = env
        self.planning_horizon = self.env.action_configs["planning_horizon"]
        self.num_steps = None

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        self.num_steps = int(self.planning_horizon / self.env.scenario.dt)

        return observation

    def step(self, action):
        """
        step function which steps the environment for the amount of steps specified
        :param action: current action
        """
        reward_list = []
        for _ in range(self.num_steps):
            next_state, reward, done, info = self.env.step(action)
            reward_list.append(reward)
            if done:
                return next_state, np.mean(reward_list), done, info
            else:
                continue

        return next_state, np.mean(reward_list), done, info



import pickle

import commonroad_dc.pycrcc as pycrcc
import numpy as np
from commonroad.scenario.scenario import ScenarioID
from commonroad.scenario.trajectory import CustomState
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object

from commonroad_rl.gym_commonroad.action.vehicle import ContinuousVehicle
from commonroad_rl.gym_commonroad.observation import TrafficSignObservation, ObservationCollector
from commonroad_rl.gym_commonroad.utils.scenario_io import restore_scenario
from commonroad_rl.gym_commonroad.utils.scenario import parse_map_name
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.path import *


xml_path = os.path.join(resource_root("test_traffic_sign"))
pickle_path = os.path.join(output_root("test_traffic_sign"), "pickles")

def prepare_for_traffic_sign_test():

    pickle_xml_scenarios(
        input_dir=xml_path,
        output_dir=pickle_path
    )

    # Traffic Sign observation settings
    traffic_sign_observation = TrafficSignObservation(
        configs={"traffic_sign_configs":
                     {"observe_stop_sign": True,
                      "observe_yield_sign": True,
                      "observe_priority_sign": True,
                      "observe_right_of_way_sign": True,
                      }})

    # specify resource path
    meta_scenario_path = os.path.join(pickle_path, "meta_scenario")
    problem_path = os.path.join(pickle_path, "problem")

    # Load meta scenario and problem dict
    filename = "DEU_AAH-3_33000_T-1.pickle"

    meta_scenario_reset_dict_path = os.path.join(meta_scenario_path, "meta_scenario_reset_dict.pickle")
    with open(meta_scenario_reset_dict_path, "rb") as f:
        meta_scenario_reset_dict = pickle.load(f)

    # Load scenarios and problems
    fn = os.path.join(problem_path, filename)
    with open(fn, "rb") as f:
        problem_dict = pickle.load(f)

    # Set scenario and problem
    scenario_id = ScenarioID.from_benchmark_id(os.path.basename(fn).split(".")[0], "2020a")
    map_id = parse_map_name(scenario_id)
    reset_config = meta_scenario_reset_dict[map_id]
    scenario = restore_scenario(reset_config["meta_scenario"], problem_dict["obstacle"], scenario_id)

    connected_lanelet_dict = meta_scenario_reset_dict[map_id]["connected_lanelet_dict"]

    lanelet_polygons = [(l.lanelet_id, l.convert_to_polygon()) for l in scenario.lanelet_network.lanelets]
    lanelet_polygons_sg = pycrcc.ShapeGroup()
    for l_id, poly in lanelet_polygons:
        lanelet_polygons_sg.add_shape(create_collision_object(poly))

    return traffic_sign_observation, scenario, connected_lanelet_dict, lanelet_polygons, lanelet_polygons_sg


traffic_sign_observation, scenario_ts, connected_lanelet_dict_ts, lanelet_polygons_ts, lanelet_polygons_sg_ts = prepare_for_traffic_sign_test()


@pytest.mark.parametrize(
    ("ego_position", "traffic_sign_expected", "distance_expected"),
    [
        (np.array([51.3018, -17.9640]),  # Previous result calculated using wrong method or by calculating distance to
                                         # wrong (not closest) signs by hand
         ["stop_sign"],
         [7.7294],  # Used to be 5.8765
         ),
        (np.array([51.3108, -30.]),
         ["yield_sign"],
         [5.2505],  # Used to be 14.6937
         ),
        (np.array([84.0947, -55.1586]),
         ["priority_sign"],
         [17.0867]  # Used to be 14.8328
         ),
        (np.array([52.5147, -20.9587]),
         ["right_of_way_sign"],
         [1.0976],  # Used to be 0.0
         ),
    ],
)
@module_test
@functional
def test_get_traffic_sign_on_lanelets(ego_position, traffic_sign_expected, distance_expected):
    """
    Test for traffic sign observation
    """
    # set ego vehicle
    ego_state = CustomState(**{"time_step": 0, "position": ego_position, "orientation": 0, "velocity": 0})
    ego_vehicle = ContinuousVehicle({"vehicle_type": 2, "vehicle_model": 2})
    ego_vehicle.reset(ego_state, dt=0.1)

    # find ego lanelet here because multiple observations need it
    ego_lanelet_ids = ObservationCollector.sorted_lanelets_by_state(scenario_ts, ego_state,
                                                                    lanelet_polygons_ts, lanelet_polygons_sg_ts)
    ego_lanelet_id = ego_lanelet_ids[0]

    ego_lanelet = scenario_ts.lanelet_network.find_lanelet_by_id(ego_lanelet_id)

    local_ccosy, _ = ObservationCollector.get_local_curvi_cosy(scenario_ts.lanelet_network,
                                                               ego_lanelet_id,
                                                               None,
                                                               max_lane_merge_range=1000.)
    # observe traffic sign on ego vehicle lanelet
    traffic_sign_observation.observe(scenario_ts, ego_vehicle, ego_lanelet, local_ccosy)

    traffic_sign_observed = []
    traffic_sign_distance = []
    if traffic_sign_observation.observation_dict["stop_sign"]:
        traffic_sign_observed.append("stop_sign")
    if traffic_sign_observation.observation_dict["yield_sign"]:
        traffic_sign_observed.append("yield_sign")
    if traffic_sign_observation.observation_dict["priority_sign"]:
        traffic_sign_observed.append("priority_sign")
    if traffic_sign_observation.observation_dict["right_of_way_sign"]:
        traffic_sign_observed.append("right_of_way_sign")

    # calculate the longitudinal distance from ego vehicle to traffic sign
    if len(traffic_sign_observed) != 0:
        for each_sign in traffic_sign_observed:
            dis = each_sign + "_distance_long"
            traffic_sign_distance.append(traffic_sign_observation.observation_dict[dis][0])
    print(traffic_sign_observed, traffic_sign_distance)

    # Check against ground truth
    assert traffic_sign_observed == traffic_sign_expected
    assert np.allclose(traffic_sign_distance, distance_expected, atol=0.001)





"""
Module tests of the module gym_commonroad
"""
import copy
import os
import glob
import random
import timeit
import numpy as np
from commonroad.scenario.scenario import ScenarioID
from stable_baselines.common.env_checker import check_env
from commonroad.common.file_reader import CommonRoadFileReader
from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_object
from commonroad_rl.gym_commonroad import *
from commonroad_rl.gym_commonroad.utils.scenario import parse_map_name
from commonroad_rl.tests.common.marker import *
from commonroad_rl.tests.common.non_functional import function_to_string
from commonroad_rl.tests.common.path import resource_root, output_root
from commonroad_rl.tools.pickle_scenario.xml_to_pickle import pickle_xml_scenarios

resource_path = resource_root("test_gym_commonroad")
pickle_xml_scenarios(
    input_dir=os.path.join(resource_path),
    output_dir=os.path.join(resource_path, "pickles")
)

meta_scenario_path = os.path.join(resource_path, "pickles", "meta_scenario")
problem_path = os.path.join(resource_path, "pickles", "problem")

output_path = output_root("test_gym_commonroad")
visualization_path = os.path.join(output_path, "visualization")


@pytest.mark.parametrize(("num_of_checks", "test_env", "play"),
                         [(15, False, False),
                          (15, False, True),
                          (15, True, False),
                          (15, True, True)])
@module_test
@functional
def test_check_env(num_of_checks, test_env, play):
    # Run more circles of checking to search for sporadic issues
    for idx in range(num_of_checks):
        print(f"Checking progress: {idx + 1}/{num_of_checks}")
        env = gym.make("commonroad-v1", meta_scenario_path=meta_scenario_path, train_reset_config_path=problem_path,
                       test_reset_config_path=problem_path, visualization_path=visualization_path, test_env=False,
                       play=False, )
        check_env(env)


@pytest.mark.parametrize(("reward_type"),
                         [("hybrid_reward"),
                          ("sparse_reward")])
@module_test
@functional
def test_step(reward_type):
    env = gym.make("commonroad-v1",
                   meta_scenario_path=meta_scenario_path,
                   train_reset_config_path=problem_path,
                   test_reset_config_path=problem_path,
                   visualization_path=visualization_path,
                   reward_type=reward_type)
    env.reset()
    done = False
    while not done:
        action = env.action_space.sample()
        obs, reward, done, info = env.step(action)


@module_test
@functional
def test_reset_env_with_scenario():
    from commonroad.scenario.scenario import Scenario
    from commonroad.scenario.obstacle import DynamicObstacle
    from commonroad.planning.planning_problem import PlanningProblem
    from commonroad_dc.collision.collision_detection.pycrcc_collision_dispatch import create_collision_checker

    env_kwargs = {"vehicle_params": {"vehicle_model": 2},
                  "cache_navigators": True,
                  "action_configs": {"continuous_collision_checking": False}}
    env_reset = gym.make("commonroad-v1",
                         meta_scenario_path=meta_scenario_path,
                         train_reset_config_path=problem_path,
                         test_reset_config_path=problem_path, **env_kwargs)
    env_step = copy.deepcopy(env_reset)

    xml_fn = sorted(glob.glob(os.path.join(resource_path, "*.xml")))[0]
    scenario, planning_problem_set = CommonRoadFileReader(xml_fn).open(lanelet_assignment=True)
    planning_problem = list(planning_problem_set.planning_problem_dict.values())[0]

    observations_reset = env_reset.reset(scenario=scenario, planning_problem=planning_problem)
    observations_step = env_step.reset(scenario=scenario, planning_problem=planning_problem)
    assert np.allclose(observations_reset, observations_step)

    collision_checker_step = create_collision_checker(scenario)
    observation_keys = []
    for key, value in env_reset.observation_dict.items():
        observation_keys.extend([key]*value.size)
    obs_distance_goal_long_advance_idx = observation_keys.index("distance_goal_long_advance")
    obs_distance_goal_lat_advance_idx = observation_keys.index("distance_goal_lat_advance")
    obs_distance_goal_long_idx = observation_keys.index("distance_goal_long")
    obs_distance_goal_lat_idx = observation_keys.index("distance_goal_lat")
    obs_distance_goal_time = observation_keys.index("distance_goal_time")
    obs_remaining_steps = observation_keys.index("remaining_steps")

    for i in range(1, 50):
        action = np.array([0.2, 1.0])
        _ = env_reset.step(action)
        observations_step, _, done, info = env_step.step(action)
        # env_step.ego_action.vehicle.update_collision_object(create_convex_hull=False)
        collision_step = collision_checker_step.collide(env_step.ego_action.vehicle.collision_object)

        obs_distance_goal_long = observations_reset[obs_distance_goal_long_idx]
        obs_distance_goal_lat = observations_reset[obs_distance_goal_lat_idx]

        # create new scenario with next step
        next_scenario = Scenario(dt=scenario.dt, scenario_id=ScenarioID("tmp"))
        next_scenario.add_objects(scenario.lanelet_network)
        for obstacle in scenario.dynamic_obstacles:
            obstacle_state = copy.deepcopy(obstacle.state_at_time(i))
            if obstacle_state is None:
                continue
            obstacle_state.time_step = 0
            next_obstacle = DynamicObstacle(obstacle.obstacle_id,
                                            obstacle.obstacle_type,
                                            obstacle.obstacle_shape,
                                            obstacle_state,
                                            prediction=None)
            next_scenario.add_objects(next_obstacle)
        next_scenario.assign_obstacles_to_lanelets(time_steps=[0])
        collision_checker_reset = create_collision_checker(next_scenario)
        ego_next_state = env_reset.ego_action.vehicle.state
        ego_next_state.time_step = 0
        ego_next_state.slip_angle = 0.

        observations_reset = env_reset.reset(scenario=next_scenario, planning_problem=PlanningProblem(
            planning_problem.planning_problem_id,
            ego_next_state,
            planning_problem.goal
        ))

        collision_reset = collision_checker_reset.collide(env_reset.ego_action.vehicle.collision_object)

        if collision_reset != collision_step:
            # find obstacle that causes collision
            for dynamic_obstacle in scenario.dynamic_obstacles:
                co = create_collision_object(dynamic_obstacle)
                if co.collide(env_step.ego_action.vehicle.collision_object):
                    obstacle_id = dynamic_obstacle.obstacle_id
                    break
            import commonroad_dc.pycrcc as pycrcc
            obstacle_reset = next_scenario.obstacle_by_id(obstacle_id)
            obstacle_state_reset = obstacle_reset.state_at_time(0)
            co_reset_obstacle = pycrcc.RectOBB(obstacle_reset.obstacle_shape.length / 2,
                                               obstacle_reset.obstacle_shape.width / 2,
                                               obstacle_state_reset.orientation,
                                               obstacle_state_reset.position[0],
                                               obstacle_state_reset.position[1])
            print(co_reset_obstacle.collide(env_reset.ego_action.vehicle.collision_object))

            obstacle_step = scenario.obstacle_by_id(obstacle_id)
            obstacle_state_step = obstacle_step.state_at_time(i)
            co_step_obstacle = pycrcc.RectOBB(obstacle_step.obstacle_shape.length / 2,
                                               obstacle_step.obstacle_shape.width / 2,
                                               obstacle_state_step.orientation,
                                               obstacle_state_step.position[0],
                                               obstacle_state_step.position[1])
            # print(co_step_obstacle.collide(env_step.ego_action.vehicle.collision_object))
            # print(co_step_obstacle.collide(env_reset.ego_action.vehicle.collision_object))

            from commonroad.visualization.mp_renderer import MPRenderer
            renderer = MPRenderer()
            co_step_obstacle.draw(renderer, draw_params={"facecolor": "blue"})
            co_reset_obstacle.draw(renderer, draw_params={"facecolor": "black"})
            env_reset.ego_action.vehicle.collision_object.draw(renderer, draw_params={"facecolor": "red"})
            env_step.ego_action.vehicle.collision_object.draw(renderer, draw_params={"facecolor": "green"})

            renderer.render(show=True)

        assert collision_reset == collision_step == \
               env_step.observation_dict['is_collision'] == env_reset.observation_dict['is_collision'], \
            f"step={i} " \
            f"ego_state_step={env_step.ego_action.vehicle.state}, ego_state_reset={env_reset.ego_action.vehicle.state}"

        # fix goal distance advance
        observations_reset[obs_distance_goal_long_advance_idx] = \
            abs(obs_distance_goal_long) - abs(observations_reset[obs_distance_goal_long_idx])
        observations_reset[obs_distance_goal_lat_advance_idx] = \
            abs(obs_distance_goal_lat) - abs(observations_reset[obs_distance_goal_lat_idx])
        observations_reset[obs_remaining_steps] -= i
        observations_reset[obs_distance_goal_time] += i

        # a = np.where(np.abs(observations_step-observations_reset) > 1e-8)
        assert np.allclose(observations_reset, observations_step), f"{env_reset.observation_dict['is_collision']}, " \
                                                                   f"{env_step.observation_dict['is_collision']}, " \
                                                                   f"collision_reset={collision_reset}, " \
                                                                   f"collision_step={collision_step}"

        if not np.allclose(observations_reset, observations_step):
            print(f"ego_state: env_reset={env_reset.ego_action.vehicle.state}, env_step={env_step.ego_action.vehicle.state}")
            keys = []
            idxs = np.where(np.abs(observations_step-observations_reset) > 1e-8)
            for idx in idxs[0]:
                keys.append(observation_keys[idx])
                print(f"{observation_keys[idx]}: observation_reset={observations_reset[idx]}, observation_step={observations_step[idx]}")
            raise ValueError


@pytest.mark.parametrize(("reward_type"),
                         [("sparse_reward"),
                          ("hybrid_reward")])
@module_test
@functional
def test_observation_order(reward_type):
    env = gym.make("commonroad-v1", meta_scenario_path=meta_scenario_path, train_reset_config_path=problem_path,
                   test_reset_config_path=problem_path, flatten_observation=False)

    # set random seed to make the env choose the same planning problem
    random.seed(0)
    obs_dict = env.reset()

    # collect observation in other format
    env = gym.make("commonroad-v1", meta_scenario_path=meta_scenario_path, train_reset_config_path=problem_path,
                   test_reset_config_path=problem_path, flatten_observation=True)

    # seed needs to be reset before function call
    random.seed(0)
    obs_flatten = env.reset()
    obs_flatten_exp = np.zeros(env.observation_space.shape)

    # flatten the dictionary observation
    index = 0
    for obs_dict_value in obs_dict.values():
        size = np.prod(obs_dict_value.shape)
        obs_flatten_exp[index: index + size] = obs_dict_value.flat
        index += size

    # compare 2 observation
    assert np.allclose(obs_flatten_exp, obs_flatten), "Two observations don't have the same order"


@pytest.mark.parametrize(("reward_type"),
                         [("sparse_reward"),
                          ("hybrid_reward")])
@module_test
@nonfunctional
def test_step_time(reward_type):
    # Define reference time
    reference_time = 15.0

    def measurement_setup():
        import gym
        import numpy as np

        env = gym.make("commonroad-v1", meta_scenario_path="{meta_scenario_path}",
                       train_reset_config_path="{problem_path}",
                       test_reset_config_path="{problem_path}", visualization_path="{visualization_path}",
                       reward_type="{reward_type}", )
        env.reset()
        action = np.array([0.0, 0.0])

    def measurement_code(env, action):
        env.step((action))

    setup_str = function_to_string(measurement_setup)
    code_str = function_to_string(measurement_code)

    times = timeit.repeat(setup=setup_str, stmt=code_str, repeat=1, number=1000)
    min_time = np.amin(times)

    # TODO: Set exclusive CPU usage for this thread, because other processes influence the result  # assert
    #  average_time < reference_time, f"The step is too slow, average time was {average_time}"
